# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division

############################################################################
#
#   Copyright (C) 2008-2015
#    Christian Kohlöffel
#    Vinzenz Schulz
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

from __future__ import absolute_import
from __future__ import division

from dxf2gcode.core.point import Point
from dxf2gcode.core.arcgeo import ArcGeo
from dxf2gcode.core.linegeo import LineGeo
from dxf2gcode.dxfimport.classes import PointsClass, ContourClass



class GeoentPolyline:
    def __init__(self, Nr=0, caller=None):
        self.Typ = 'Polyline'
        self.Nr = Nr
        self.Layer_Nr = 0
        self.geo = []
        self.length = 0

        # Lesen der Geometrie
        # Read the geometry
        self.Read(caller)

    def __str__(self):
        # how to print the object
        string = "\nTyp: Polyline" +\
                 "\nNr: %i" % self.Nr +\
                 "\nLayer Nr: %i" % self.Layer_Nr +\
                 "\nNr. of Lines: %i" % len(self.geo) +\
                 "\nlength: %0.3f" % self.length

        return string

    def reverse(self):
        """
        reverse()
        """
        self.geo.reverse()
        for geo in self.geo:
            geo.reverse()

    def App_Cont_or_Calc_IntPts(self, cont, points, i, tol, warning):
        """
        App_Cont_or_Calc_IntPts()
        """
        if abs(self.length) < tol:
            pass

        # Hinzuf�gen falls es keine geschlossene Polyline ist
        # Add if it is not a closed polyline
        elif self.geo[0].Ps.within_tol(self.geo[-1].Pe, tol):
            self.analyse_and_opt()
            cont.append(ContourClass(len(cont), 1, [[i, 0]], self.length))
        else:
            points.append(PointsClass(point_nr=len(points), geo_nr=i,
                                      Layer_Nr=self.Layer_Nr,
                                      be=self.geo[0].Ps,
                                      en=self.geo[-1].Pe, be_cp=[], en_cp=[]))

        return warning

#        if abs(self.length)>tol:
#            points.append(PointsClass(point_nr=len(points),geo_nr=i,\
#                                      Layer_Nr=self.Layer_Nr,\
#                                      be=self.geo[0].Ps,
#                                      en=self.geo[-1].Pe,be_cp=[],en_cp=[]))
#        else:
#            showwarning("Short Polyline Elemente", ("Length of Line geometrie too short!"\
#                                               "\nLenght must be greater than tolerance."\
#                                               "\nSkipping Line Geometrie"))

    def analyse_and_opt(self):
        """
        analyse_and_opt()
        """
        summe = 0

        # Richtung in welcher der Anfang liegen soll (unten links)
        # Direction of the top (lower left) ???
        Popt = Point(-1e3, -1e6)

        # Calculation of the alignment after Gaussian-Elling
        # Positive value means CW, negative value indicates CCW
        # closed polygon
        for Line in self.geo:
            summe += Line.Ps.x * Line.Pe.y - Line.Pe.x * Line.Ps.y

        if summe > 0.0:
            self.reverse()

        # Suchen des kleinsten Startpunkts von unten Links X zuerst (Muss neue Schleife sein!)
        # Find the smallest starting point from bottom left X (Must be new loop!)
        min_distance = self.geo[0].Ps.distance(Popt)
        min_geo_nr = 0
        for geo_nr in range(1, len(self.geo)):
            if self.geo[geo_nr].Ps.distance(Popt) < min_distance:
                min_distance = self.geo[geo_nr].Ps.distance(Popt)
                min_geo_nr = geo_nr

        # Kontur so anordnen das neuer Startpunkt am Anfang liegt
        # Order Contour so the new starting point is at the beginning
        self.geo = self.geo[min_geo_nr:len(self.geo)] + self.geo[0:min_geo_nr]

    def Read(self, caller):
        """
        Read()
        """
        # Assign short name
        lp = caller.line_pairs
        e = lp.index_both(0, "SEQEND", caller.start + 1) + 1

        # Assign layer
        s = lp.index_code(8, caller.start + 1)
        self.Layer_Nr = caller.Get_Layer_Nr(lp.line_pair[s].value)

        # Ps=None for the first point
        Ps = None

        # Polyline flag
        s_temp = lp.index_code(70, s + 1, e)
        if s_temp is None:
            PolyLineFlag = 0
        else:
            PolyLineFlag = int(lp.line_pair[s_temp].value)
            s = s_temp

        # print("PolylineFlag: %i" %PolyLineFlag)

        while 1:  # and s is not None:
            s = lp.index_both(0, "VERTEX", s + 1, e)
            if s == None:
                break

            # X Value
            s = lp.index_code(10, s + 1, e)
            x = float(lp.line_pair[s].value)

            # Y Value
            s = lp.index_code(20, s + 1, e)
            y = float(lp.line_pair[s].value)
            Pe = Point(x, y)

            # Bulge
            bulge = 0

            e_vertex = lp.index_both(0, "VERTEX", s + 1, e)
            if e_vertex is None:
                e_vertex = e

            s_temp = lp.index_code(42, s + 1, e_vertex)
            # print('stemp: %s, e: %s, next 10: %s' %(s_temp,e,lp.index_both(0,"VERTEX",s+1,e)))
            if s_temp is not None:
                bulge = float(lp.line_pair[s_temp].value)
                s = s_temp

            # Vertex flag (bit-coded); default is 0; 1 = Closed; 128 = Plinegen
            s_temp = lp.index_code(70, s + 1, e_vertex)
            if s_temp is None:
                VertexFlag = 0
            else:
                VertexFlag = int(lp.line_pair[s_temp].value)
                s = s_temp

            # print("Vertex Flag: %i" %PolyLineFlag)

            # Assign the geometries for the Polyline
            if VertexFlag != 16:
                if Ps is not None:
                    if next_bulge == 0:
                        self.geo.append(LineGeo(Ps=Ps, Pe=Pe))
                    else:
                        # self.geo.append(LineGeo(Ps=Ps,Pe=Pe))
                        # print bulge
                        self.geo.append(self.bulge2arc(Ps, Pe, next_bulge))

                    # L�nge drauf rechnen wenns eine Geometrie ist
                    # Wenns Ldnge count on it is a geometry ???
                    self.length += self.geo[-1].length

                # Der Bulge wird immer f�r den und den n�chsten Punkt angegeben
                # The bulge is always given for the next point
                next_bulge = bulge
                Ps = Pe

        # It is a closed polyline
        if PolyLineFlag == 1 and len(self.geo) > 0:
            # print("sollten �bereinstimmen: %s, %s" %(Ps,Pe))
            if next_bulge == 0:
                self.geo.append(LineGeo(Ps=Ps, Pe=self.geo[0].Ps))
            else:
                self.geo.append(self.bulge2arc(Ps, self.geo[0].Ps, next_bulge))
            # L�nge drauf rechnen wenns eine Geometrie ist
            # Wenns Ldnge count on it is a geometry ???
            self.length += self.geo[-1].length

        # Neuen Startwert f�r die n�chste Geometrie zur�ckgeben
        # New starting value for the next geometry
        caller.start = e

    def get_start_end_points(self, direction=0):
        """
        get_start_end_points()
        """
        if not direction:
            punkt, angle = self.geo[0].get_start_end_points(direction)
        else:
            punkt, angle = self.geo[-1].get_start_end_points(direction)
        return punkt, angle

    def bulge2arc(self, Ps, Pe, bulge):
        """
        bulge2arc()
        """
        c = (1 / bulge - bulge) / 2

        # Berechnung des Mittelpunkts (Formel von Mickes!)
        # Calculation of the center (Micke's formula)
        O = Point((Ps.x + Pe.x - (Pe.y - Ps.y) * c) / 2,
                  (Ps.y + Pe.y + (Pe.x - Ps.x) * c) / 2)

        # Radius = Distance between the centre and Ps
        r = O.distance(Ps)
        # Kontrolle ob beide gleich sind (passt ...)
        # Check if they are equal (fits ...)
        # r=O.distance(Pe)

        # Unterscheidung f�r den �ffnungswinkel.
        # Distinction for the opening angle. ???
        if bulge > 0:
            return ArcGeo(Ps=Ps, Pe=Pe, O=O, r=r)
        else:
            arc = ArcGeo(Ps=Pe, Pe=Ps, O=O, r=r)
            arc.reverse()
            return arc



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2010-2014
#    Christian Kohl�ffel
#    Jean-Paul Schouwstra
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

import dxf2gcode.globals.constants as c
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QGridLayout, QTextBrowser
from PyQt5.QtGui import QIcon, QPixmap, QTextCursor
from PyQt5 import QtCore

import logging

logger = logging.getLogger("Gui.AboutDialog")


class AboutDialog(QDialog):
    def __init__(self, title="Test", message="Test Text"):
        super(AboutDialog, self).__init__()

        self.title = title
        self.message = message

        self.initUI()

    def initUI(self):
        """
        initUI()
        """

        vbox = QVBoxLayout(self)
        grid1 = QGridLayout()
        grid1.setSpacing(10)

        self.text = QTextBrowser()
        self.text.setReadOnly(True)
        self.text.setOpenExternalLinks(True)
        self.text.append(self.message)
        self.text.moveCursor(QTextCursor.Start)
        self.text.ensureCursorVisible()

        vbox.addWidget(self.text)

        self.setLayout(vbox)
        self.setMinimumSize(550, 450)
        self.resize(550, 600)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        self.setWindowTitle(self.title)
        iconWT = QIcon()
        iconWT.addPixmap(QPixmap(":images/DXF2GCODE-001.ico"),
                         QIcon.Normal, QIcon.Off)
        self.setWindowIcon(QIcon(iconWT))

        self.exec_()



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2014-2015
#    Robert Lichtenberger
#    Jean-Paul Schouwstra
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

from __future__ import absolute_import
from __future__ import division

from copy import deepcopy
from math import pi

from dxf2gcode.core.point import Point
from dxf2gcode.core.boundingbox import BoundingBox


class HoleGeo(object):

    """
    HoleGeo represents drilling holes.
    """

    def __init__(self, Ps):
        """
        Standard Method to initialise the HoleGeo
        """
        self.Ps = Ps
        self.length = -1

        self.abs_geo = None

    def __deepcopy__(self, memo):
        return HoleGeo(deepcopy(self.Ps, memo))

    def __str__(self):
        """
        Standard method to print the object
        @return: A string
        """
        return "\nHoleGeo at (%s) " % self.Ps

    def save_v1(self):
        return "\nHoleGeo at (%s) " % self.Ps.save_v1()

    def calc_bounding_box(self, radius=1):
        """
        Calculated the BoundingBox of the geometry and saves it into self.BB
        @param radius: The Radius of the HoleGeo to be used for BoundingBox
        """

        Ps = Point(x=self.Ps.x - radius, y=self.Ps.y - radius)
        Pe = Point(x=self.Ps.x + radius, y=self.Ps.y + radius)

        self.BB = BoundingBox(Ps=Ps, Pe=Pe)

    def reverse(self):
        """
        Reverses the direction.
        """
        pass

    def make_abs_geo(self, parent=None):
        """
        Generates the absolute geometry based on itself and the parent. This
        is done for rotating and scaling purposes
        """
        Ps = self.Ps.rot_sca_abs(parent=parent)

        self.abs_geo = HoleGeo(Ps)

    def get_start_end_points(self, start_point, angles=None):
        if angles is None:
            return self.Ps
        elif angles:
            return self.Ps, 0
        else:
            return self.Ps, Point(0, -1) if start_point else Point(0, -1)

    def make_path(self, caller, drawHorLine):

        radius = caller.parentLayer.tool_diameter / 2
        self.calc_bounding_box(radius=radius)

        segments = 30
        Ps = self.Ps.get_arc_point(0, radius)

        for i in range(1, segments + 1):
            ang = i * 2 * pi / segments
            Pe = self.Ps.get_arc_point(ang, radius)
            drawHorLine(caller, Ps, Pe)
            Ps = Pe

    def isHit(self, caller, xy, tol):
        """
        This function returns true if the nearest point between the two geometries is within the square of the 
        given tolerance
        @param caller: This is the calling entities (only used in holegeo)
        @param xy: The point which shall be used to determine the distance
        @tol: The tolerance which is used for Hit testing.
        """

        return abs(self.ps.distance(xy) - caller.parentLayer.getToolRadius()) < tol

    def Write_GCode(self, PostPro):
        """
        Writes the GCODE for a Hole.
        @param PostPro: The PostProcessor instance to be used
        @return: Returns the string to be written to a file.
        """
        return PostPro.make_print_str("(Drilled hole)%nl")



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2009-2016
#    Christian Kohlöffel
#    Jean-Paul Schouwstra
#    Xavier Izard
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

from __future__ import absolute_import

import os
import sys
import pprint
import logging

from collections import OrderedDict
from configobj import ConfigObj, flatten_errors
from validate import Validator
import dxf2gcode.globals.globals as g
from dxf2gcode.globals.d2gexceptions import *
from dxf2gcode.gui.configwindow import *

import dxf2gcode.globals.constants as c
from PyQt5 import QtCore


logger = logging.getLogger("Core.Config")

CONFIG_VERSION = "9.10"
"""
version tag - increment this each time you edit CONFIG_SPEC

compared to version number in config file so
old versions are recognized and skipped"
"""

# Paths change whether platform is Linux or Windows
if "linux" in sys.platform.lower() or "openbsd" in sys.platform.lower():
    #Declare here the path that are specific to Linux
    IMPORT_DIR = "~/Documents"
    OUTPUT_DIR = "~/Documents"
    PDFTOPS_CMD = "/usr/local/bin/pdftops"
    PSTOEDIT_CMD = "/usr/local/bin/pstoedit"
else:
    #Declare here the path that are specific to Windows
    IMPORT_DIR = "D:/Eclipse_Workspace/DXF2GCODE/trunk/dxf"
    OUTPUT_DIR = "D:"
    # Xpdf Tools from http://www.xpdfreader.com/download.html
    PDFTOPS_CMD = "C:/xpdf-tools-win-4.00/bin64/pdftops.exe"
    # pstoedit from https://sourceforge.net/projects/pstoedit/
    # and https://sourceforge.net/projects/ghostscript/ dependency
    # if gs in not on the PATH it might be necessary to add -gs option
    # to psedit in Option->Configuration->Software config->pstoedit
    # e.g.  -gs, C:/Program Files (x86)/gs/gs9.09/bin/gswin32c.exe
    PSTOEDIT_CMD = "C:/Program Files/pstoedit/pstoedit.exe"

"""
HOWTO declare a new variable in the config file:
1) Choose the appropriate section and add the variable in the CONFIG_SPEC string below
(Note: the CONFIG_SPEC is used to set and check the configfile "config.cfg")

2) Set it's default value, the min/max values (if applicable) and a comment above the variable's name
(Important note: the min/max values and the comment are directly used in the configuration window, so carefully set them!)
Example of correct declaration:
    [MySection]
    # Drag angle is used to blah blah blah ...
    drag_angle = float(min = 0, max = 360, default = 20)

3) If you want the setting to appear in the configuration window, fill the cfg_widget_def variable, using the _same_ names as in the CONFIG_SPEC
Example of declaration correlated to the above one:
    'MySection':
    {
        'drag_angle': CfgDoubleSpinBox('Drag angle (in degrees):'),
    }
(Note: the list of available types for the configuration window can be found in the "configwindow.py" file)
"""
"""
ATTENTION:
_Don't_ split the long comments lines in CONFIG_SPEC!
The comments line are used as "QWhatsThis" in the config window.
Any new line in the CONFIG_SPEC is reproduced in the QWhatsThis (intented behaviour to allow creating paragraphs)
ATTENTION
"""
CONFIG_SPEC = str('''
#  Section and variable names must be valid Python identifiers
#      do not use whitespace in names

# do not edit the following section name:
    [Version]
    # do not edit the following value:
    config_version = string(default = "''' +
    str(CONFIG_VERSION) + '")\n' +
    '''
    [Paths]
    # By default look for DXF files in this directory.
    import_dir = string(default = "''' + IMPORT_DIR + '''")

    # Export generated gcode by default to this directory.
    output_dir = string(default = "''' + OUTPUT_DIR + '''")

    [Filters]
    # pstoedit is an external tool to convert PDF files to PS (postscript) files, which can be further processed by pstoedit tool.
    pdftops_cmd = string(default = "''' + PDFTOPS_CMD + '''")
    pdftops_opt = list(default = list())

    # pstoedit is an external tool to import *.ps (postscript) files and convert them to DXF, in order to import them in dxf2gcode.
    pstoedit_cmd = string(default = "''' + PSTOEDIT_CMD + '''")
    pstoedit_opt = list(default = list('-dt', '-f', 'dxf:-mm'))

    [Axis_letters]
    ax1_letter = string(min = 1, default = "X")
    ax2_letter = string(min = 1, default = "Y")
    ax3_letter = string(min = 1, default = "Z")

    [Plane_Coordinates]
    axis1_start_end = float(default = 0)
    axis2_start_end = float(default = 0)

    [Depth_Coordinates]
    # Third axis' coordinate at which it can do rapid move.
    axis3_retract = float(default = 15.0)
    # Third axis' margin for which it needs to do a slow move.
    axis3_safe_margin = float(default = 3.0)
    # The top third axis' coordinate of the workpiece.
    axis3_start_mill_depth = float(default = 0.0)
    # Relative depth for each cut (third axis' coordinate will be decreased by this value at each step).
    axis3_slice_depth = float(default = -1.5)
    # Relative final third axis' depth.
    axis3_mill_depth = float(default = -3.0)

    [Feed_Rates]
    f_g1_plane = float(default = 400)
    f_g1_depth = float(default = 150)

    [General]
    # Enable 3D representation of the piece (requires PyQt5 and OpenGL)
    mode3d = boolean(default = False)
    # Write output to stdout (console), instead of a file. May be used to interface directly with Linux CNC, for example.
    write_to_stdout = boolean(default = False)
    # When enabled, the shapes that are disabled are still shown on the graphic view.
    show_disabled_paths = boolean(default = True)
    # When enabled, export path is live updated on the graphic view.
    live_update_export_route = boolean(default = False)
    # Divide the lines in 2 parts, in order to start the cutting in the middle of a line (usefull for cutter compensation)
    split_line_segments = boolean(default = False)
    # Automatically enable cutter compensation for all the shapes (G41 & G42)
    automatic_cutter_compensation = boolean(default = False)
    # Machine types supported: milling; lathe; drag_knife
    machine_type = option('milling', 'lathe', 'drag_knife', default = 'milling')
    # The unit used for all values in this file
    tool_units = option('mm', 'in', default = 'mm')

    [Cutter_Compensation]
    # If not checked, DXF2GCODE will create a virtual path for G41 and G42 command. And output will be set to G40; i.e. it will create the path that normally your machine would create with it's cutter compensation.
    done_by_machine = boolean(default = True)


    [Drag_Knife_Options]
    # drag_angle: if angle of movement exceeds this angle (in degrees), the tool retracts to dragDepth (The dragDepth is given by axis3_slice_depth parameter).
    # This parameter depends on the knife that you are using. A bigger knife cannot make small corners like a smaller knife. You will simply break your knife or destroy your working piece. Now, if the angle your knife has to make is bigger than this angle it will move to a different depth (a less deep position) such that the knife will experience less resistance but still has some (otherwise it will not change its angle at all, whence DRAG knife).
    drag_angle = float(min = 0, max = 360, default = 20)

    [Route_Optimisation]
    # If enabled, it will by default check the TSP for all the shapes in the treeview.
    # If disabled and no shape is checked for TSP optimisation in the listbox, the export order will be as defined in the listbox.
    default_TSP = boolean(default = False)

    # Path optimizer behaviour:
    # - CONSTRAIN_ORDER_ONLY: fixed Shapes and optimized Shapes can be mixed. Only order of fixed shapes is kept
    # - CONSTRAIN_PLACE_AFTER: optimized Shapes are always placed after any fixed Shape
    TSP_shape_order = option('CONSTRAIN_ORDER_ONLY', 'CONSTRAIN_PLACE_AFTER', default = 'CONSTRAIN_ORDER_ONLY')
    # This is a value of how much it should deviate the order with each iteration. The higher the value the more you allow alterations.
    mutation_rate = float(min = 0, max = 1, default = 0.95)
    # Number of people the population has for path optimization (values higher than 200 can make everything slow).
    max_population = integer(min = 0, max = 10000, default = 20)
    # Maximum number of iterations that will be done. This is internally also calculated, based on the number of shapes to optimize.
    # Values higher than 10000 can take really long to solve the TSP and are not recommended.
    max_iterations = integer(min = 1, max = 1000000, default = 300)
    # Different methods to initialize the population for the TSP optimizer.
    # - Ordered will start with the defined one in the listbox
    # - Random just random
    # - Heuristic will search the nearest neighbors and starts with the resulting order.
    begin_art = option('ordered', 'random', 'heuristic', default = 'heuristic')

    [Import_Parameters]
    # Tolerance at which similar points will be interpreted as similar
    point_tolerance = float(min = 0, max = 1, default = 0.001)
    # Types of check performed during spline conversion:
    # 1: Checking for Nurbs degree (polygon degree) and similar knots consistence
    # 2: Checking for Nurbs degree (polygon degree) and similar control points
    # 3: Performes check 1 and check 2
    spline_check = integer(min = 1, max = 3, default = 3)
    # This is the tolerance which is used to fit the converted lines and arc segments to the converted NURBS.
    fitting_tolerance = float(min = 0, max = 1, default = 0.001)
    # If checked, the elements (shape, ...) which are part of a block will be inserted on the layer that belongs to the block (even though the elements might be defined on a different layers)
    insert_at_block_layer = boolean(default = False)

    # These settings are intented to be used in the DXF file:
    # - By using MILL: as a prefix to your layer name, you can define milling parameters by using one of the following identifiers.
    # - Example of a layer name: MILL: 1 Md: 2 Sd: 2 FeedXY: 400 FeedZ: 200
    #   (This will cut shapes on the layer 2 mm deep (in one pass, since Sd == Md) using 400 mm / minute speed for X/Y movement and 200 mm / minute for Z movement)
    [Layer_Options]
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    id_float_separator = string(default = ":")

    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    mill_depth_identifiers = list(default = list('MillDepth', 'Md', 'TiefeGesamt', 'Tg'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    slice_depth_identifiers = list(default = list('SliceDepth', 'Sd', 'TiefeZustellung', 'Tz'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    start_mill_depth_identifiers = list(default = list('StartMillDepth', 'SMd', 'StartTiefe', 'St'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    retract_identifiers = list(default = list('RetractHeight', 'Rh', 'Freifahrthoehe', 'FFh'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    safe_margin_identifiers = list(default = list('SafeMargin', 'Sm', 'Sicherheitshoehe', 'Sh'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    f_g1_plane_identifiers = list(default = list('FeedXY', 'Fxy', 'VorschubXY', 'Vxy', 'F'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    f_g1_depth_identifiers = list(default = list('FeedZ', 'Fz', 'VorschubZ', 'Vz'))

    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    tool_nr_identifiers = list(default = list('ToolNr', 'Tn', 'T', 'WerkzeugNummer', 'Wn'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    tool_diameter_identifiers = list(default = list('ToolDiameter', 'Td', 'WerkzeugDurchmesser', 'Wd'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    spindle_speed_identifiers = list(default = list('SpindleSpeed', 'Drehzahl', 'RPM', 'UPM', 'S'))
    # To be used in the DXF layer name. See DXF2GCODE' wiki for more information.
    start_radius_identifiers = list(default = list('StartRadius', 'Sr'))

    # Tools table: define here the tools used for milling:
    # - name: this is the number of the tool, it will be used directly in the GCODE (eg 20 for tool T20)
    # - diameter: diameter of the tool
    # - speed: spindle speed in rpm
    # - start_radius: start radius for tool compensation when using G41 / G42 moves
    [Tool_Parameters]
    [[1]]
    diameter = float(default = 2.0)
    speed = float(default = 6000)
    start_radius = float(default = 0.2)

    [[2]]
    diameter = float(default = 2.0)
    speed = float(default = 6000.0)
    start_radius = float(default = 1.0)

    [[10]]
    diameter = float(default = 10.0)
    speed = float(default = 6000.0)
    start_radius = float(default = 2.0)

    [[__many__]]
    diameter = float(default = 3.0)
    speed = float(default = 6000)
    start_radius = float(default = 3.0)

    # Define here custom GCODE actions:
    # - name: this is the unique name of the action
    # - gcode: the text that will be inserted in the final program (each new line is also translated as a new line in the output file)
    # Custom actions can be inserted in the program by using right-click contextual menu on the treeview.
    [Custom_Actions]
    [[__many__]]
    gcode = string(default = "(change subsection name and insert your custom GCode here. Use triple quote to place the code on several lines)")

    [Logging]
    # Logging to textfile is disabled by default
    logfile = string(default = "")

    # This really goes to stderr
    console_loglevel = option('DEBUG', 'INFO', 'WARNING', 'ERROR','CRITICAL', default = 'CRITICAL')

    # Log levels are, in increasing importance: DEBUG; INFO; WARNING; ERROR; CRITICAL
    # Log events with importance >= loglevel are logged to the corresponding output
    file_loglevel = option('DEBUG', 'INFO', 'WARNING', 'ERROR','CRITICAL', default = 'DEBUG')

    # Logging level for the message window
    window_loglevel = option('DEBUG', 'INFO', 'WARNING', 'ERROR','CRITICAL', default = 'INFO')

''').splitlines()
""" format, type and default value specification of the global config file"""


class MyConfig(object):
    """
    This class hosts all functions related to the Config File.
    """
    def __init__(self):
        """
        initialize the varspace of an existing plugin instance
        init_varspace() is a superclass method of plugin
        """

        self.folder = os.path.join(g.folder, c.DEFAULT_CONFIG_DIR)
        self.filename = os.path.join(self.folder, 'config' + c.CONFIG_EXTENSION)

        self.version_mismatch = '' # no problem for now
        self.default_config = False # whether a new name was generated
        self.var_dict = dict()
        self.spec = ConfigObj(CONFIG_SPEC, interpolation=False, list_values=False, _inspec=True)

        # try:
        self.load_config()
        self.update_config()


        # The following settings won't be modified after a change in the configuration window.
        # If a setting need to be updated when the configuration changes, move it to the update_config() function
        self.mode3d = self.vars.General['mode3d']

        self.machine_type = self.vars.General['machine_type']
        self.fitting_tolerance = self.vars.Import_Parameters['fitting_tolerance']
        self.point_tolerance = self.vars.Import_Parameters['point_tolerance']

        self.metric = 0 if self.vars.General['tool_units'] == 'in' else 1   # Standard if no other unit is determined while importing
        self.tool_units = self.vars.General['tool_units'] # store the initial tool_units (we don't want it to change until software restart)
        self.tool_units_metric = 0 if self.vars.General['tool_units'] == 'in' else 1

        # except Exception, msg:
        #     logger.warning(self.tr("Config loading failed: %s") % msg)
        #     return False

    def tr(self, string_to_translate):
        """
        Translate a string using the QCoreApplication translation framework
        @param string_to_translate: a unicode string
        @return: the translated unicode string if it was possible to translate
        """
        return str(QtCore.QCoreApplication.translate('MyConfig',
                                                           string_to_translate))

    def update_config(self):
        """
        Call this function each time the self.var_dict is updated (eg when the configuration window changes some settings)
        """
        # convenience - flatten nested config dict to access it via self.config.sectionname.varname
        self.vars = DictDotLookup(self.var_dict)
        # add here any update needed for the internal variables of this class

    def make_settings_folder(self):
        """Create settings folder if necessary"""
        try:
            os.makedirs(self.folder)
        except OSError:
            pass

    def load_config(self):
        """Load Config File"""
        
        logger.info(self.filename)
        
        if os.path.isfile(self.filename):
            try:
                # file exists, read & validate it
                self.var_dict = ConfigObj(self.filename, configspec=CONFIG_SPEC)
                _vdt = Validator()
                result = self.var_dict.validate(_vdt, preserve_errors=True)
                validate_errors = flatten_errors(self.var_dict, result)

                if validate_errors:
                    logger.error(self.tr("errors reading %s:") % self.filename)

                for entry in validate_errors:
                    section_list, key, error = entry
                    if key is not None:
                        section_list.append(key)
                    else:
                        section_list.append('[missing section]')
                    section_string = ', '.join(section_list)
                    if not error:
                        error = self.tr('Missing value or section.')
                    logger.error(section_string + ' = ' + error)

                if validate_errors:
                    raise BadConfigFileError("syntax errors in config file")

                # check config file version against internal version
                if CONFIG_VERSION:
                    fileversion = self.var_dict['Version']['config_version']  # this could raise KeyError

                    if fileversion != CONFIG_VERSION:
                        raise VersionMismatchError(fileversion, CONFIG_VERSION)

            except VersionMismatchError:
                #raise VersionMismatchError(fileversion, CONFIG_VERSION)
                # version mismatch flag, it will be used to display an error.
                self.version_mismatch = self.tr("The configuration file version ({0}) doesn't match the software expected version ({1}).\n\nYou have to delete (or carefully edit) the configuration file \"{2}\" to solve the problem.").format(fileversion, CONFIG_VERSION, self.filename)

            except Exception as inst:
                logger.error(inst)
                (base, ext) = os.path.splitext(self.filename)
                badfilename = base + c.BAD_CONFIG_EXTENSION
                logger.debug(self.tr("trying to rename bad cfg %s to %s") % (self.filename, badfilename))
                try:
                    os.rename(self.filename, badfilename)
                except OSError as e:
                    logger.error(self.tr("rename(%s,%s) failed: %s") % (self.filename, badfilename, e.strerror))
                    raise
                else:
                    logger.debug(self.tr("renamed bad varspace %s to '%s'") % (self.filename, badfilename))
                    self.create_default_config()
                    self.default_config = True
                    logger.debug(self.tr("created default varspace '%s'") % self.filename)
            else:
                self.default_config = False
                # logger.debug(self.dir())
                # logger.debug(self.tr("created default varspace '%s'") % self.filename)
                # logger.debug(self.tr("read existing varspace '%s'") % self.filename)
        else:
            self.create_default_config()
            self.default_config = True
            logger.debug(self.tr("created default varspace '%s'") % self.filename)

        self.var_dict.main.interpolation = False  # avoid ConfigObj getting too clever

    def create_default_config(self):
        # check for existing setting folder or create one
        self.make_settings_folder()

        # derive config file with defaults from spec
        self.var_dict = ConfigObj(configspec=CONFIG_SPEC)
        _vdt = Validator()
        self.var_dict.validate(_vdt, copy=True)
        self.var_dict.filename = self.filename
        self.var_dict.write()

    def save_varspace(self):
        """Saves Variables space"""
        self.var_dict.filename = self.filename
        self.var_dict.write()

    def print_vars(self):
        """Prints Variables"""
        print("Variables:")
        for k, v in self.var_dict['Variables'].items():
            print(k, "=", v)

    def makeConfigWidgets(self):
        """
        Build the configuration widgets and store them into a dictionary.
        The structure of the dictionnary must match the structure of the configuration file. The names of the keys must be identical to those used in the configfile.
        If a name is declared in the configfile but not here, it simply won't appear in the config window (the config_version for example must not be modified by the user, so it is not declared here)
        """
        coordinate_unit = self.tr(" mm") if self.tool_units_metric else self.tr(" in")
        speed_unit = self.tr(" mm/min") if self.tool_units_metric else self.tr(" IPS")
        cfg_widget_def = OrderedDict([
            ('__section_title__', {
                # This section is only used for assigning titles to the keys of the dictionnary (= name of the sections used in the config file).
                # This name is displayed in the tabs of the configuration window ; if no name is provided, the key name is used. A same title may be used for several keys : in this case, all the items that belongs to this section are regrouped into the same tab.
                # Note: the title may be specified in the section itselt too, using special entry "'__section_title__': Title to use"
                'Tool_Parameters': self.tr("Tools table"),
                'Custom_Actions' : self.tr("Custom actions")}),
            ('Paths', OrderedDict([
                ('__section_title__', self.tr("Software config")),
                ('import_dir', CfgLineEdit(self.tr('By default look for DXF files in:'))),
                ('output_dir', CfgLineEdit(self.tr('By default export generated G-Code to:')))
            ])),
            ('Filters', OrderedDict([
                ('__section_title__', self.tr("Software config")),
                ('__subtitle__', CfgSubtitle("pdftops")),
                ('pdftops_cmd', CfgLineEdit(self.tr('Location of executable:'))),
                ('pdftops_opt', CfgListEdit(self.tr('Command-line options:'), ',')),

                ('__subtitle2__', CfgSubtitle("pstoedit")),
                ('pstoedit_cmd', CfgLineEdit(self.tr('Location of executable:'))),
                ('pstoedit_opt', CfgListEdit(self.tr('Command-line options:'), ','))
            ])),
            ('General', OrderedDict([
                ('__section_title__', self.tr("General settings")),
                ('mode3d', CfgCheckBox(self.tr('3D mode (requires OpenGL - restart needed)'))),
                ('show_disabled_paths', CfgCheckBox(self.tr('Display disabled paths (default)'))),
                ('live_update_export_route', CfgCheckBox(self.tr('Live update export route (default)'))),
                ('__subtitle2__', CfgSubtitle(self.tr("Milling"))),
                ('split_line_segments', CfgCheckBox(self.tr('Split line segments (default)'))),
                ('automatic_cutter_compensation', CfgCheckBox(self.tr('Automatic cutter compensation (default)'))),
                ('machine_type', CfgComboBox(self.tr('Machine type (default):'))),
                ('tool_units', CfgComboBox(self.tr('Configuration values use the unit (restart needed):'))),
                ('__subtitle3__', CfgSubtitle(self.tr("Output"))),
                ('write_to_stdout', CfgCheckBox(self.tr('Export the G-Code to stdout (instead of a file)')))
            ])),
            ('Axis_letters', OrderedDict([
                ('__section_title__', self.tr("Machine config")),
                ('__subtitle__', CfgSubtitle(self.tr("Axes' identifiers"))),
                ('ax1_letter', CfgLineEdit(self.tr('First axis:'))),
                ('ax2_letter', CfgLineEdit(self.tr('Second axis:'))),
                ('ax3_letter', CfgLineEdit(self.tr('Third axis:')))
            ])),
            ('Plane_Coordinates', OrderedDict([
                ('__section_title__', self.tr("Machine config")),
                ('__subtitle__', CfgSubtitle(self.tr("Start and end's coordinate"))),
                ('axis1_start_end', CfgDoubleSpinBox(self.tr("First axis:"), coordinate_unit)),
                ('axis2_start_end', CfgDoubleSpinBox(self.tr("Second axis:"), coordinate_unit))
            ])),
            ('Depth_Coordinates', OrderedDict([
                ('__section_title__', self.tr("Machine config")),
                ('__subtitle__', CfgSubtitle(self.tr("Third axis' defaults"))),
                ('axis3_retract', CfgDoubleSpinBox(self.tr("Retraction coordinate:"), coordinate_unit)),
                ('axis3_safe_margin', CfgDoubleSpinBox(self.tr("Safety margin:"), coordinate_unit)),
                ('axis3_start_mill_depth', CfgDoubleSpinBox(self.tr("Workpiece origin coordinate:"), coordinate_unit)),
                ('axis3_slice_depth', CfgDoubleSpinBox(self.tr("Slice depth:"), coordinate_unit)),
                ('axis3_mill_depth', CfgDoubleSpinBox(self.tr("Final mill depth:"), coordinate_unit))
            ])),
            ('Feed_Rates', OrderedDict([
                ('__section_title__', self.tr("Machine config")),
                ('__subtitle__', CfgSubtitle(self.tr("G1 feed rates"))),
                ('f_g1_plane', CfgDoubleSpinBox(self.tr('First and second axis (2D plane):'), speed_unit)),
                ('f_g1_depth', CfgDoubleSpinBox(self.tr('Third axis:'), speed_unit))
            ])),
            ('Cutter_Compensation', OrderedDict([
                ('__section_title__', self.tr("Output settings")),
                ('__subtitle__', CfgSubtitle(self.tr("Cutter compensation"))),
                ('done_by_machine', CfgCheckBox(self.tr('Cutter compensation is done by machine (check box if machine reconizes G41 and G42 commands / uncheck otherwise)')))
            ])),
            ('Drag_Knife_Options', OrderedDict([
                ('__section_title__', self.tr("Output settings")),
                ('__subtitle__', CfgSubtitle(self.tr("Drag Knife options"))),
                ('drag_angle', CfgDoubleSpinBox(self.tr('Retract to drag depth ("Third axis slice depth") if angle exceeds:'), u'°'))#u for unicode
            ])),
            ('Route_Optimisation', OrderedDict([
                ('__section_title__', self.tr("Output settings")),
                ('default_TSP', CfgCheckBox(self.tr("By default enable the TSP optimization for ordering shapes (TSP = Travelling Salesman Problem)"))),
                ('__subtitle__', CfgSubtitle(self.tr("TSP optimizer"))),
                ('TSP_shape_order', CfgComboBox(self.tr('Optimizer behaviour:'))),
                ('mutation_rate', CfgDoubleSpinBox(self.tr('Mutation rate:'))),
                ('max_population', CfgSpinBox(self.tr('Max population:'))),
                ('max_iterations', CfgSpinBox(self.tr('Max iterations:'))),
                ('begin_art', CfgComboBox(self.tr('Starting point:')))
            ])),
            ('Import_Parameters', OrderedDict([
                ('__section_title__', self.tr("Output settings")),
                ('point_tolerance', CfgDoubleSpinBox(self.tr('DXF default import point tolerance:'), '', None, None, 5)),
                ('spline_check', CfgSpinBox(self.tr('DXF import spline check:'))),
                ('fitting_tolerance', CfgDoubleSpinBox(self.tr('DXF default import fit tolerance:'), '', None, None, 5)),
                ('insert_at_block_layer', CfgCheckBox(self.tr('insert elements which are part of a block to layer where the block is inserted')))
            ])),
            ('Layer_Options', OrderedDict([
                ('__section_title__', self.tr("Automatic tool config")),
                ('id_float_separator', CfgLineEdit(self.tr('Separator between identifier and value:'))),
                ('__subtitle__', CfgSubtitle(self.tr("Parameters' identifiers"))),
                ('mill_depth_identifiers', CfgListEdit(self.tr('Final mill depth:'), ',')),
                ('slice_depth_identifiers', CfgListEdit(self.tr('Infeed depth:'), ',')),
                ('start_mill_depth_identifiers', CfgListEdit(self.tr('Workpiece top:'), ',')),
                ('retract_identifiers', CfgListEdit(self.tr('Retraction area:'), ',')),
                ('safe_margin_identifiers', CfgListEdit(self.tr('Safety margin:'), ',')),
                ('f_g1_plane_identifiers', CfgListEdit(self.tr('G1 feed rate for first and second axis (2D plane):'), ',')),
                ('f_g1_depth_identifiers', CfgListEdit(self.tr('G1 feed rate for third axis:'), ',')),
                ('tool_nr_identifiers', CfgListEdit(self.tr('Tool number:'), ',')),
                ('tool_diameter_identifiers', CfgListEdit(self.tr('Tool diameter:'), ',')),
                ('spindle_speed_identifiers', CfgListEdit(self.tr('Spindle speed:'), ',')),
                ('start_radius_identifiers', CfgListEdit(self.tr('Start radius (cutter compensation):'), ','))
            ])),
            ('Tool_Parameters', CfgTableToolParameters(self.tr('Define the tools here:'))),
            ('Custom_Actions', CfgTableCustomActions(self.tr('Define here custom GCODE that can be inserted anywhere in the program:'))),
            ('Logging', OrderedDict([
                ('__section_title__', self.tr("Software config")),
                ('logfile', CfgLineEdit(self.tr('File used for logging (restart needed):'))),
                ('console_loglevel', CfgComboBox(self.tr('On stderr console log messages with importance minimal to level (restart needed):'))),
                ('file_loglevel', CfgComboBox(self.tr('For log-file log messages with importance minimal to level (restart needed):'))),
                ('window_loglevel', CfgComboBox(self.tr('For message box log messages with importance minimal to level (restart needed):')))
            ]))
        ])

        return cfg_widget_def

    def update_tool_values(self):
        """
        update the tool default values depending on the unit of the drawing
        """
        if self.tool_units_metric != self.metric:
            scale = 1/25.4 if self.metric == 0 else 25.4
            for key in self.vars.Plane_Coordinates:
                self.vars.Plane_Coordinates[key] *= scale
            for key in self.vars.Depth_Coordinates:
                self.vars.Depth_Coordinates[key] *= scale
            for key in self.vars.Feed_Rates:
                self.vars.Feed_Rates[key] *= scale
            for tool in self.vars.Tool_Parameters:
                self.vars.Tool_Parameters[tool]['diameter'] *= scale
                self.vars.Tool_Parameters[tool]['start_radius'] *= scale
            self.tool_units_metric = self.metric


class DictDotLookup(object):
    """
    Creates objects that behave much like a dictionaries, but allow nested
    key access using object '.' (dot) lookups.
    """
    def __init__(self, d):
        for k in d:
            if isinstance(d[k], dict):
                self.__dict__[k] = DictDotLookup(d[k])
            elif isinstance(d[k], (list, tuple)):
                l = []
                for v in d[k]:
                    if isinstance(v, dict):
                        l.append(DictDotLookup(v))
                    else:
                        l.append(v)
                self.__dict__[k] = l
            else:
                self.__dict__[k] = d[k]

    def __getitem__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]

    def __setitem__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value

    def __iter__(self):
        return iter(self.__dict__.keys())

    def __repr__(self):
        return pprint.pformat(self.__dict__)

# if __name__ == '__main__':
#     cfg_data = eval("""{
#         'foo' : {
#             'bar' : {
#                 'tdata' : (
#                     {'baz' : 1 },
#                     {'baz' : 2 },
#                     {'baz' : 3 },
#                 ),
#             },
#         },
#         'quux' : False,
#     }""")
#
#     cfg = DictDotLookup(cfg_data)
#
#     # iterate
#     for k, v in cfg.__iter__(): #foo.bar.iteritems():
#         print k, " = ", v
#
#     print "cfg=", cfg
#
#     #   Standard nested dictionary lookup.
#     print 'normal lookup :', cfg['foo']['bar']['tdata'][0]['baz']
#
#     #   Dot-style nested lookup.
#     print 'dot lookup    :', cfg.foo.bar.tdata[0].baz
#
#     print "qux=", cfg.quux
#     cfg.quux = '123'
#     print "qux=", cfg.quux
#
#     del cfg.foo.bar
#     cfg.foo.bar = 4711
#     print 'dot lookup    :', cfg.foo.bar #.tdata[0].baz



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2008-2015
#    Christian Kohlöffel
#    Vinzenz Schulz
#    Jean-Paul Schouwstra
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################


class EntityContent(object):
    def __init__(self, nr, name, parent, p0, pb, sca, rot):
        """
        @param p0: The Starting Point to plot (Default x=0 and y=0)
        @param bp: The Base Point to insert the geometry and base for rotation
        (Default is also x=0 and y=0)
        @param sca: The scale of the basis function (default =1)
        @param rot: The rotation of the geometries around base (default =0)
        """
        self.nr = nr
        self.name = name
        self.parent = parent
        self.children = []
        self.p0 = p0
        self.pb = pb
        self.sca = sca
        self.rot = rot

    def __str__(self):
        return "\nEntityContent" +\
               "\nnr :      %i" % self.nr +\
               "\nname:     %s" % self.name +\
               "\nchildren: %s" % self.children +\
               "\np0:       %s" % self.p0 +\
               "\npb:       %s" % self.pb +\
               "\nsca:      %s" % self.sca +\
               "\nrot:      %s" % self.rot

    def append(self, child):
        self.children.append(child)



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2009-2015
#    Michael Haberler
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

"""
Container for global variables accessible to all classes
"""

from __future__ import absolute_import

import os
import sys
import gettext
import locale

import dxf2gcode.globals.constants as constants


# logger instance, see http://docs.python.org/library/logging.html
# once set, use as logger.error("foo")
logger = None

# Config instance
config = None

# Folder of the main instance
folder = None

window = None

quiet = False

#-------------------------------------

# determine Platform
platform = ""
if os.name == "posix" and sys.platform == "darwin":
    platform = "mac"


# Language support
#
langs = []  # list of supported languages

# figure default language
lc, encoding = locale.getdefaultlocale()

if lc:
    langs = [lc]    # if there's one, use as default

language = os.environ.get('LANGUAGE', None)
if language:
    """language comes back something like en_CA:en_US:en_GB:en
    on linuxy systems, on Win32 it's nothing, so we need to
    split it up into a list"""
    langs += language.split(":")

"""Now add on to the back of the list the translations that we
know that we have, our defaults"""
langs += []

"""Now langs is a list of all of the languages that we are going
to try to use.  First we check the default, then what the system
told us, and finally the 'known' list"""

gettext.bindtextdomain(constants.APPNAME,
                       os.path.realpath(os.path.dirname(sys.argv[0])))
gettext.textdomain(constants.APPNAME)
# Get the language to use
trans = gettext.translation(constants.APPNAME, localedir='languages',
                            languages=langs, fallback=True)
trans.install()



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2012-2015
#    Xavier Izard
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################


class CustomGCode(object):
    """
    This class contains a "custom gcode" object. Custom GCode objects are part
    of a layer (layercontent.py) and are used to insert custom GCode into the
    generated file.
    Custom GCodes are defined in the config file

    @purpose: store user defined GCode
    """
    def __init__(self, name, nr, gcode, parentLayer):
        """
        Standard method to initialize the class
        @param name: the name of the GCode, as defined in the config file
        @param gcode: the user defined gcode
        @param parent: The parent layer Class of the shape
        """
        self.name = name
        self.nr = nr
        self.gcode = gcode
        self.parentLayer = parentLayer
        self.disabled = False
        self.send_to_TSP = False  # Never optimize path for CustomGCode
        self.selected = False

    def __str__(self):
        """
        Standard method to print the object
        @return: A string
        """
        return "\nCustomGCode" +\
               "\nname:  %s" % self.name +\
               "\nnr:    %i" % self.nr +\
               "\ngcode: %s" % self.gcode

    def setDisable(self, flag=False):
        """
        Function to modify the disable property
        @param flag: The flag to enable or disable Selection
        """
        self.disabled = flag

    def isDisabled(self):
        """
        Returns the state of self.disabled
        """
        return self.disabled

    def setSelected(self, flag=False):
        self.selected = flag

    def isSelected(self):
        return self.selected

    def Write_GCode(self, PostPro):
        """
        This method returns the string to be exported for this custom gcode
        @param PostPro: this is the Postprocessor class including the methods to export
        """
        return self.gcode



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2008-2015
#    Christian Kohlöffel
#    Vinzenz Schulz
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

from __future__ import absolute_import
from __future__ import division

from math import sin, cos, pi, ceil
from dxf2gcode.core.linegeo import LineGeo
from dxf2gcode.core.arcgeo import ArcGeo
from dxf2gcode.core.point import Point


class BiarcClass(object):
    """
    BiarcClass Class for the generation of Biarc Section. It is used for the
    Ellipse fitting and the Nurbs conversion to Line and Arc segments.
    """
    def __init__(self, Ps=Point(), tan_a=0.0, Pb=Point, tan_b=0.0, min_r=1e-6):
        """
        Std. method to initialise the class.
        @param Ps: Start Point for the Biarc
        @param tan_a: Tangent of the Start Point
        @param Pb: End Point of the Biarc
        @param tan_b: Tangent of the End Point
        @param min_r: The minimum radius of a arc section.
        """
        min_len = 1e-12       # Min Abstand f�r doppelten Punkt / Minimum clearance for double point
        min_alpha = 1e-4      # Winkel ab welchem Gerade angenommen wird inr rad / Angle for which it is assumed straight inr rad
        max_r = 5e3           # Max Radius ab welchem Gerade angenommen wird (5m) / Max radius is assumed from which line (5m)
        min_r = min_r         # Min Radius ab welchem nichts gemacht wird / Min radius beyond which nothing is done

        self.Ps = Ps
        self.tan_a = tan_a
        self.Pb = Pb
        self.tan_b = tan_b
        self.l = 0.0
        self.shape = None
        self.geos = []
        self.k = 0.0

        # Errechnen der Winkel, L�nge und Shape
        # Calculate the angle, length and shape
        norm_angle, self.l = self.calc_normal(self.Ps, self.Pb)

        alpha, beta, self.theta, self.shape = self.calc_diff_angles(norm_angle,
                                                                    self.tan_a,
                                                                    self.tan_b,
                                                                    min_alpha)

        if self.l < min_len:
            self.shape = "Zero"

        elif self.shape == "LineGeo":
            # Erstellen der Geometrie
            # Create the geometry
            self.geos.append(LineGeo(self.Ps, self.Pb))
        else:
            # Berechnen der Radien, Mittelpunkte, Zwichenpunkt
            # Calculate the radii, midpoints Zwichenpunkt
            r1, r2 = self.calc_r1_r2(self.l, alpha, beta, self.theta)

            if abs(r1) > max_r or abs(r2) > max_r:
                # Erstellen der Geometrie
                # Create the geometry
                self.shape = "LineGeo"
                self.geos.append(LineGeo(self.Ps, self.Pb))
                return

#             elif abs(r1) < min_r or abs(r2) < min_r:
#                 self.shape = "Zero"
#                 return

            O1, O2, k = self.calc_O1_O2_k(r1, r2, self.tan_a, self.theta)

            # Berechnen der Start und End- Angles f�r das drucken
            # Calculate the start and end angles for the print
            s_ang1, e_ang1 = self.calc_s_e_ang(self.Ps, O1, k)
            s_ang2, e_ang2 = self.calc_s_e_ang(k, O2, self.Pb)

            # Berechnen der Richtung und der Extend
            # Calculate the direction and extent
            dir_ang1 = (tan_a - s_ang1) % (-2 * pi)
            dir_ang1 -= ceil(dir_ang1 / pi) * (2 * pi)

            dir_ang2 = (tan_b - e_ang2) % (-2 * pi)
            dir_ang2 -= ceil(dir_ang2 / pi) * (2 * pi)

            # Erstellen der Geometrien
            # Create the geometries
            self.geos.append(ArcGeo(Ps=self.Ps, Pe=k, O=O1, r=r1,
                                    s_ang=s_ang1, e_ang=e_ang1, direction=dir_ang1))
            self.geos.append(ArcGeo(Ps=k, Pe=self.Pb, O=O2, r=r2,
                                    s_ang=s_ang2, e_ang=e_ang2, direction=dir_ang2))

    def __str__(self):
        s = "\nBiarc Shape: %s" % self.shape +\
            "\nPs : %s; Tangent: %0.3f" % (self.Ps, self.tan_a) +\
            "\nPb : %s; Tangent: %0.3f" % (self.Pb, self.tan_b) +\
            "\ntheta: %0.3f, l: %0.3f" % (self.theta, self.l)
        for geo in self.geos:
            s += str(geo)
        return s

    def calc_O1_O2_k(self, r1, r2, tan_a, theta):
        """
        calc_O1_O2_k()
        """
        # print("r1: %0.3f, r2: %0.3f, tan_a: %0.3f, theta: %0.3f" %(r1,r2,tan_a,theta))
        # print("N1: x: %0.3f, y: %0.3f" %(-sin(tan_a), cos(tan_a)))
        # print("V: x: %0.3f, y: %0.3f" %(-sin(theta+tan_a),cos(theta+tan_a)))

        O1 = Point(self.Ps.x - r1 * sin(tan_a),
                   self.Ps.y + r1 * cos(tan_a))
        k = Point(self.Ps.x + r1 * (-sin(tan_a) + sin(theta + tan_a)),
                  self.Ps.y + r1 * (cos(tan_a) - cos(tan_a + theta)))
        O2 = Point(k.x + r2 * (-sin(theta + tan_a)),
                   k.y + r2 * (cos(theta + tan_a)))
        return O1, O2, k

    def calc_normal(self, Ps, Pb):
        """
        calc_normal()
        """
        norm_angle = Ps.norm_angle(Pb)
        l = Ps.distance(Pb)
        return norm_angle, l

    def calc_diff_angles(self, norm_angle, tan_a, tan_b, min_alpha):
        """
        calc_diff_angles()
        """
        # print("Norm angle: %0.3f, tan_a: %0.3f, tan_b %0.3f" %(norm_angle,tan_a,tan_b))
        alpha = (norm_angle - tan_a)
        beta = (tan_b - norm_angle)
        alpha, beta = self.limit_angles(alpha, beta)

        if alpha * beta > 0.0:
            shape = "C-shaped"
            theta = alpha
        elif abs(alpha - beta) < min_alpha:
            shape = "LineGeo"
            theta = alpha
        else:
            shape = "S-shaped"
            theta = (3 * alpha - beta) / 2

        return alpha, beta, theta, shape

    def limit_angles(self, alpha, beta):
        """
        limit_angles()
        """
        # print("limit_angles: alpha: %s, beta: %s" %(alpha,beta))
        if alpha < -pi:
            alpha += 2 * pi
        if alpha > pi:
            alpha -= 2 * pi
        if beta < -pi:
            beta += 2 * pi
        if beta > pi:
            beta -= 2 * pi
        while pi < (alpha - beta):
            alpha -= 2 * pi
        while -pi > (alpha - beta):
            alpha += 2 * pi
        # print("   -->>       alpha: %s, beta: %s" %(alpha,beta))
        return alpha, beta

    def calc_r1_r2(self, l, alpha, beta, theta):
        """
        calc_r1_r2()
        """
        # print("alpha: %s, beta: %s, theta: %s" %(alpha,beta,theta))
        r1 = (l / (2 * sin((alpha + beta) / 2)) *
              sin((beta - alpha + theta) / 2) / sin(theta / 2))
        r2 = (l / (2 * sin((alpha + beta) / 2)) *
              sin((2 * alpha - theta) / 2) / sin((alpha + beta - theta) / 2))
        return r1, r2

    def calc_s_e_ang(self, P1, O, P2):
        """
        calc_s_e_ang()
        """
        s_ang = O.norm_angle(P1)
        e_ang = O.norm_angle(P2)
        return s_ang, e_ang

    def get_biarc_fitting_error(self, Pt):
        """
        get_biarc_fitting_error()
        """
        # Query in which segment of the circle the point is:
        w1 = self.geos[0].O.norm_angle(Pt)
        if (w1 >= min([self.geos[0].s_ang, self.geos[0].e_ang]))and\
           (w1 <= max([self.geos[0].s_ang, self.geos[0].e_ang])):
            diff = self.geos[0].O.distance(Pt) - abs(self.geos[0].r)
        else:
            diff = self.geos[1].O.distance(Pt) - abs(self.geos[1].r)
        return abs(diff)



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2015-2016
#    Xavier Izard
#    Jean-Paul Schouwstra
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

"""
@purpose: build a configuration window on top of ConfigObj configfile module.

It aims to be generic and reusable for many configuration windows.

*** Basic usage ***
1) Let's say that your ConfigObj spec file is declared as is:
CONFIG_SPEC = str('''
    [MySection]
    # Comment for the variable below
    my_variable = float(min = 0, max = 360, default = 20)
    ''').splitlines()

2) Declare the corresponding dictionnary
config_widget_dict = OrderedDict([
        ('MySection', OrderedDict([
            ('__section_title__', "My Title"),
            ('__subtitle__', CfgSubtitle(self.tr("My possible subtitle"))),
            ('my_variable', CfgDoubleSpinBox("My parameter description")),
            ('my_variable2', CfgDoubleSpinBox("My parameter description2")),
            ('NextSubtitleX', CfgSubtitle(self.tr("My not needed second subtitle"))),
            ('my_variable3', CfgDoubleSpinBox("My parameter description3"))
        ])),
    ])

The my __subtitle__ is not restricted to be placed under __section_title__ it can be placed on any line and
it is also not restricted to be named like that. You can even leave it out. If you do it's replaced by a line.
If you place it on a different line (with the name:__subtitle__), this subsection does not start with horizontal bar.

3) Instanciate the config window:
config_window = ConfigWindow(config_widget_dict, var_dict, configspec, self) #See ConfigObj for var_dict & configspec
config_window.finished.connect(self.updateConfiguration) #Optional signal to know when the config has changed

*** List of graphical elements currently supported ***
 - CfgSubtitle(): subtitle - just for displaying a bar with some text
 - CfgCheckBox(): a basic (possibly tristate) checkbox
 - CfgSpinBox(): a spinbox for int values
 - CfgDoubleSpinBox(): a spinbox for float values
 - CfgLineEdit(): a text input (1 line)
 - CfgListEdit(): a text list input (1 line)
 - CfgTextEdit(): a text input (multiple lines)
 - CfgComboBox(): a drop-down menu for selecting options
 - CfgTable(): a 2D table with editable text entries
 - CfgTableCustomActions(): specific module based on CfgTable(), for storing custom GCODE
 - CfgTableToolParameters(): specific module based on CfgTable(), for storing mill tools
"""

from __future__ import absolute_import

import logging
from collections import OrderedDict

from dxf2gcode.globals.helperfunctions import toInt, toFloat, str_encode, qstr_encode

import dxf2gcode.globals.constants as c

from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QMessageBox, QVBoxLayout, QHBoxLayout, QLayout, QFrame, \
    QLabel, QLineEdit, QTextEdit, QCheckBox, QSpinBox, QDoubleSpinBox, QComboBox, QTableWidget, QTableWidgetItem, \
    QPushButton, QAbstractItemView, QWidget, QSizePolicy, QListWidget, QStackedWidget, QSplitter
from PyQt5.QtGui import QIcon, QPixmap, QValidator, QRegExpValidator
from PyQt5.QtCore import QLocale, QRegExp
from PyQt5 import QtCore

from dxf2gcode.gui.popupdialog import PopUpDialog

logger = logging.getLogger("Gui.ConfigWindow")


class ConfigWindow(QDialog):
    Applied = QDialog.Accepted + QDialog.Rejected + 1 #Define a result code that is different from accepted and rejected

    """Main Class"""
    def __init__(self, definition_dict, config = None, configspec = None, parent = None, title = "Configuration"):
        """
        Initialization of the Configuration window. ConfigObj must be instanciated before this one.
        @param definition_dict: the dict that describes our window
        @param config: data readed from the configfile. This dict is created by ConfigObj module.
        @param configspec: specifications of the configfile. This variable is created by ConfigObj module.
        """
        QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        iconWT = QIcon()
        iconWT.addPixmap(QPixmap(":images/DXF2GCODE-001.ico"), QIcon.Normal, QIcon.Off)
        self.setWindowIcon(QIcon(iconWT))

        self.edition_mode = False #No editing in progress for now

        self.cfg_window_def = definition_dict #This is the dict that describes our window
        self.var_dict = config #This is the data from the configfile (dictionary created by ConfigObj class)
        self.configspec = configspec #This is the specifications for all the entries defined in the config file

        #There is no config file selector for now, so no callback either
        self.selector_change_callback = None
        self.selector_add_callback = None
        self.selector_remove_callback = None
        self.selector_duplicate_callback = None

        #Create the vars for the optional configuration's file selector
        self.cfg_file_selector = None
        self.frame_file_selector = CfgBase() #For displaying the optionnal files selector widgets

        #Create the config window according to the description dict received
        self.list_items = self.createWidgetFromDefinitionDict()

        #Create 3 buttons
        self.button_box = QDialogButtonBox(QDialogButtonBox.Discard | QDialogButtonBox.Apply | QDialogButtonBox.Close)
        self.button_box.accepted.connect(self.accept) #Apply and close (currently unused)
        self.button_box.rejected.connect(self.reject) #Close
        apply_button = self.button_box.button(QDialogButtonBox.Apply)
        apply_button.clicked.connect(self.applyChanges) #Apply button
        discard_button = self.button_box.button(QDialogButtonBox.Discard)
        discard_button.clicked.connect(self.discardChanges) #Discard button

        #Layout
        list_widget = QListWidget(parent)
        self.tab_window = QStackedWidget()
        for label, widget in self.list_items.items():
            list_widget.addItem(label)
            self.tab_window.addWidget(widget)

        list_widget.currentTextChanged.connect(self.selectionChanged)

        tab_widget = CfgBase()
        tab_box = QVBoxLayout(self)
        tab_box.addWidget(self.frame_file_selector)
        tab_box.addWidget(self.tab_window)
        tab_widget.setLayout(tab_box)

        splitter = QSplitter()
        splitter.addWidget(list_widget)
        splitter.addWidget(tab_widget)

        #Layout the 2 above widgets vertically
        v_box = QVBoxLayout(self)
        v_box.addWidget(splitter)
        v_box.addWidget(self.button_box)
        self.setLayout(v_box)

        #Populate our Configuration widget with the values from the config file
        if self.var_dict is not None and self.configspec is not None:
            self.affectValuesFromConfig(self.var_dict, self.configspec)

        #No modification in progress for now
        self.setEditInProgress(False)


    def keyPressEvent(self, event):
        """
        Reimplemented keyPressEvent() function so that we can catch and ignore the [ENTER] key
        (When pressed inside a QDialog, this key apply the changes by default)
        """
        if event.key() == QtCore.Qt.Key_Enter or event.key() == QtCore.Qt.Key_Return:
            #key [ENTER]
            event.accept() #We caught the key and we "eat" it, so it prevents its default behaviour
        else:
            #Default behaviour for all the other keys
            QDialog.keyPressEvent(self, event)


    def tr(self, string_to_translate):
        """
        Translate a string using the QCoreApplication translation framework
        @param: string_to_translate: a unicode string
        @return: the translated unicode string if it was possible to translate
        """
        return str(QtCore.QCoreApplication.translate('ConfigWindow', string_to_translate))


    def setEditInProgress(self, edit_mode):
        """
        @param edit_mode: when True, the configuration window swith to edition mode, meaning that the "Apply" and "OK" buttons are enabled, ...
        """
        editor_mode = edit_mode != False
        self.button_box.button(QDialogButtonBox.Apply).setEnabled(editor_mode)
        self.button_box.button(QDialogButtonBox.Discard).setEnabled(editor_mode)
        self.button_box.button(QDialogButtonBox.Close).setEnabled(not editor_mode)
        self.frame_file_selector.setEnabled(not editor_mode)


    def selectionChanged(self, text):
        """
        Slot called when a category is selected in the list of the config widget
        """
        self.tab_window.setCurrentWidget(self.list_items[str(text)])


    def accept(self):
        """
        Check and apply the changes, then close the config window (OK button)
        """
        ok, errors_list = self.validateConfiguration(self.cfg_window_def)
        if ok:
            self.updateConfiguration(self.cfg_window_def, self.var_dict) #Update the configuration dict according to the new settings in our config window
            QDialog.accept(self)
            logger.info('New configuration OK')
            #No more modification in progress
            self.setEditInProgress(False)
        else:
            self.displayMessageBox(errors_list)

    def applyChanges(self):
        """
        Apply changes without closing the window (allow to test some changes without reopening the config window each time)
        """
        ok, errors_list = self.validateConfiguration(self.cfg_window_def)
        if ok:
            self.updateConfiguration(self.cfg_window_def, self.var_dict) #Update the configuration dict according to the new settings in our config window
            self.setResult(ConfigWindow.Applied) #Return a result code that is different from accepted and rejected
            self.finished.emit(self.result())
            logger.info('New configuration applied')
            #No more modification in progress
            self.setEditInProgress(False)
        else:
            self.displayMessageBox(errors_list)

    def reject(self):
        """
        Reload our configuration widget with the values from the config file (=> Cancel the changes in the config window), then close the config window
        """
        self.affectValuesFromConfig(self.var_dict, self.configspec) # cannot be in the if statement - it is possible that the changeOccured was not fired
        if not self.button_box.button(QDialogButtonBox.Close).isEnabled():
            logger.info('New configuration cancelled')
        self.setEditInProgress(False)
        QDialog.reject(self)


    def discardChanges(self):
        """
        Reload our configuration widget with the values from the config file (=> Cancel the changes in the config window). Don't close the config window
        """
        self.affectValuesFromConfig(self.var_dict, self.configspec)
        logger.info('New configuration cancelled')


    def displayMessageBox(self, errors_list):
        """
        Popup a message box in order to display an error message
        @param errors_list: a string that contains all the errors
        """
        errors_list = self.tr('Please correct the following error(s):\n') + errors_list
        error_message = QMessageBox(QMessageBox.Critical, self.tr('Invalid changes'), errors_list);
        error_message.exec_()


    def setConfigSelectorCallback(self, selection_changed_callback, add_file_callback, remove_file_callback, duplicate_file_callback):
        """
        Define the functions called when the config file selector is used (respectively when a new file is selected in the combobox / when a file is added / when a file is removed)
        The ConfigWindow class is just toplevel configuration widget, it doesn't know about anything about the data, hence the callback
        """
        self.selector_change_callback = selection_changed_callback
        self.selector_add_callback = add_file_callback
        self.selector_remove_callback = remove_file_callback
        self.selector_duplicate_callback = duplicate_file_callback


    def setConfigSelectorFilesList(self, config_list, select_item = None):
        """
        Define the functions called when the config file selector is used (respectively when a new file is selected in the combobox / when a file is added / when a file is removed)
        The ConfigWindow class is just toplevel configuration widget, it doesn't know about anything about the data, hence the callback
        """
        if len(config_list) > 0:
            if self.frame_file_selector.layout() is None:
                #There is currently no file selector widget, so we create a new one
                self.cfg_file_selector = CfgComboBox("Choose configuration file:", None, None)
                button_duplicate = QPushButton(QIcon(QPixmap(":/images/layer.png")), "")
                button_duplicate.setToolTip(self.tr("Duplicate the current post-processor"))
                button_add = QPushButton(QIcon(QPixmap(":/images/list-add.png")), "")
                button_add.setToolTip(self.tr("Add a new post-processor with default values"))
                button_remove = QPushButton(QIcon(QPixmap(":/images/list-remove.png")), "")
                button_remove.setToolTip(self.tr("Remove the current post-processor"))

                #Connect the signals to call the callback when an action is done on the file selector
                if self.selector_change_callback is not None:
                    self.cfg_file_selector.combobox.currentIndexChanged[int].connect(self.selector_change_callback)
                button_duplicate.clicked.connect(self.configSelectorDuplicateFile)
                button_add.clicked.connect(self.configSelectorAddFile)
                button_remove.clicked.connect(self.configSelectorRemoveFile)

                layout_file_selector = QHBoxLayout() #For displaying the optional file's selector widget
                layout_file_selector.addWidget(self.cfg_file_selector)
                layout_file_selector.addWidget(button_duplicate)
                layout_file_selector.addWidget(button_add)
                layout_file_selector.addWidget(button_remove)
                self.frame_file_selector.setLayout(layout_file_selector)

            #Fill the combobox with the current file list
            self.cfg_file_selector.setSpec({'string_list': config_list['filename'], 'comment': ''})

            #Select the item if not None
            if select_item is not None:
                self.cfg_file_selector.setValue(select_item)

            #Load the current config
            if self.selector_change_callback is not None:
                self.selector_change_callback(self.cfg_file_selector.combobox.currentIndex())

        else:
            #Item should be a layout or a widget
            logger.warning("At least one config file must be passed to the config selector!")

            #Remove all the files from the config file selector
            if self.cfg_file_selector is not None:
                self.cfg_file_selector.setSpec({'string_list': [], 'comment': ''})


    def configSelectorDuplicateFile(self):
        """
        Function called when the "Remove configuration file" is clicked in the optional config selector zone
        """
        title = self.tr('Duplicate a configuration file')
        label = [self.tr("Enter a new filename (without extension):")]
        value = [""]
        filename_dialog = PopUpDialog(title, label, value)

        if filename_dialog.result is not None and len(filename_dialog.result[0]) > 0:
            #Call the callback function to duplicate the file
            if self.selector_duplicate_callback is not None:
                if self.selector_duplicate_callback(str(self.cfg_file_selector.getValue()), str(filename_dialog.result[0])) == False: #note: str() is needed for PyQT4
                    self.displayMessageBox(self.tr('An error occured while duplicating the file "{0}". Check that it doesn\'t already exists for example'.format(filename_dialog.result[0])))
            else:
                logger.warning("No callback defined for duplicating the file, nothing will happen!")


    def configSelectorAddFile(self):
        """
        Function called when the "Remove configuration file" is clicked in the optional config selector zone
        """
        title = self.tr('Add a configuration file')
        label = [self.tr("Enter filename (without extension):")]
        value = [""]
        filename_dialog = PopUpDialog(title, label, value)

        if filename_dialog.result is not None and len(filename_dialog.result[0]) > 0:
            if self.selector_add_callback is not None:
                if self.selector_add_callback(str(filename_dialog.result[0])) == False:
                    self.displayMessageBox(self.tr('An error occured while creating the file "{0}". Check that it doesn\'t already exists for example'.format(filename_dialog.result[0])))
            else:
                logger.warning("No callback defined for adding the file, nothing will happen!")


    def configSelectorRemoveFile(self):
        """
        Function called when the "Remove configuration file" is clicked in the optional config selector zone
        """
        confirmation_result = QMessageBox.question(self, self.tr('Delete configuration file?'), self.tr('Are you sure you want to permanently remove the file "{0}"'.format(self.cfg_file_selector.getValue())), QMessageBox.Ok | QMessageBox.Cancel);
        if confirmation_result == QMessageBox.Yes or confirmation_result == QMessageBox.Ok:
            #User has confirmed the file suppression, so let's go
            if self.selector_remove_callback is not None:
                if self.selector_remove_callback(str(self.cfg_file_selector.getValue())) == False:
                    self.displayMessageBox(self.tr('An error occured while removing the file "{0}". Remove it manually'.format(self.cfg_file_selector.getValue())))
            else:
                logger.warning("No callback defined for removing the file, nothing will happen!")


    def createWidgetFromDefinitionDict(self):
        """
        Automatically build a widget, based on dict definition of the items.
        @return: a QWidget containing all the elements of the configuration window
        """
        logger.info('Creating configuration window')
        definition = self.cfg_window_def

        tab_widgets = OrderedDict()

        #Create a dict with the sections' titles if not already defined. This dict contains sections' names as key and tabs' titles as values
        if '__section_title__' not in definition:
            definition['__section_title__'] = {}

        #Compute all the sections
        for section in definition:
            #skip the special section __section_title__
            if section == '__section_title__':
                continue

            #Create the title for the section if it doesn't already exist
            if section not in definition['__section_title__']:
                #The title for this section doesn't exist yet
                if isinstance(definition[section], dict) and '__section_title__' in definition[section]:
                    #The title for this section is defined into the section itself => we add the title to the dict containing all the titles
                    definition['__section_title__'][section] = definition[section]['__section_title__']
                else:
                    #The title for this section is not defined anywhere, so we use the section name itself as a title
                    definition['__section_title__'][section] = section.replace('_', ' ')

            #Create the tab (and the widget) for the current section, if it doesn't exist yet
            widget = None
            for widget_label in tab_widgets:
                if definition['__section_title__'][section] == widget_label:
                    widget = tab_widgets[widget_label]
                    break

            if widget is None:
                widget = QWidget()
                tab_widgets[definition['__section_title__'][section]] = widget

            #Create the tab content for this section
            self.createWidgetSubSection(definition[section], widget)

            #Add a stretch at the end of this subsection
            if widget.layout() is not None:
                widget.layout().addStretch()

        #Add a QSpacer at the bottom of each widget, so that the items are placed on top of each tab
        for widget in tab_widgets.values():
            if widget.layout() is not None:
                widget.layout().addStretch()

        return tab_widgets


    def createWidgetSubSection(self, subdefinition, section_widget):
        """
        Create the widgets that will be inserted into the tabs of the configuration window
        @param subdefinition: part of the definition dict
        @param section_widget: the widget that host the subwidgets
        @return: section_widget (for recursive call)
        """
        vertical_box = section_widget.layout()
        if vertical_box is None:
            vertical_box = QVBoxLayout()
            section_widget.setLayout(vertical_box)
        vertical_box.setSpacing(0) #Don't use too much space, it makes the option window too big otherwise

        if isinstance(subdefinition, dict):
            vertical_box.addWidget(subdefinition.get('__subtitle__', CfgSubtitle()))

        self.createWidgetSubSectionWithSubLevels(subdefinition, section_widget)


    def createWidgetSubSectionWithSubLevels(self, subdefinition, section_widget):
        """
        Create the widgets that will be inserted into the tabs of the configuration window
        @param subdefinition: part of the definition dict
        @param section_widget: the widget that host the subwidgets
        @return: section_widget (for recursive call)
        """

        vertical_box = section_widget.layout()

        if isinstance(subdefinition, dict):
            for subsection in subdefinition:
                if subsection == '__section_title__':
                    #skip the special section
                    continue

                #Browse sublevels
                self.createWidgetSubSectionWithSubLevels(subdefinition[subsection], section_widget) #Recursive call, all the nested configuration item will appear at the same level
        else:
            if isinstance(subdefinition, (QWidget, QLayout)):
                vertical_box.addWidget(subdefinition)
                if hasattr(subdefinition, 'setChangeSlot'):
                    subdefinition.setChangeSlot(self.changeOccured)
            else:
                #Item should be a layout or a widget
                logger.error("item subdefinition is incorrect")

        return section_widget


    def affectValuesFromConfig(self, config, configspec):
        """
        Affect new values for the configuration
        @param config: data readed from the configfile. This dict is created by ConfigObj module.
        @param configspec: specifications of the configfile. This variable is created by ConfigObj module.
        """
        self.var_dict = config #This is the data from the configfile (dictionary created by ConfigObj class)
        self.configspec = configspec #This is the specifications for all the entries defined in the config file

        self.setValuesFromConfig(self.cfg_window_def, self.var_dict, self.configspec)

        #No modification in progress for now
        self.setEditInProgress(False)

    def setValuesFromConfig(self, window_def, config, configspec):
        """
        This function populates the option widget with the values that come from the configuration file.
        The values from the configuration file are stored into a dictionary, we browse this dictionary to populate our window
        @param window_def: the dict that describes our window
        @param config: data readed from the configfile. This dict is created by ConfigObj module.
        @param configspec: specifications of the configfile. This variable is created by ConfigObj module.
        """
        #Compute all the sections
        for section in window_def:
            #skip the special section __section_title__
            if section == '__section_title__' or isinstance(window_def[section], CfgDummy):
                continue

            if config is not None and section in config:
                if isinstance(window_def[section], dict):
                    #Browse sublevels
                    configspec_sub = None
                    if configspec is not None and section in configspec:
                        configspec_sub = configspec[section]
                    self.setValuesFromConfig(window_def[section], config[section], configspec_sub) #Recursive call, until we find a real item (not a dictionnary with subtree)
                else:
                    if isinstance(window_def[section], (QWidget, QLayout)):
                        #assign the configuration retrieved from the configspec object of the ConfigObj
                        if configspec is not None and section in configspec:
                            window_def[section].setSpec(self.configspecParser(configspec[section], configspec.comments[section]))
                        #assign the value that was readed from the configfile
                        window_def[section].setValue(config[section])
                    else:
                        #Item should be a layout or a widget
                        logger.warning("item {0} is not a widget, can't set it's value!".format(window_def[section]))
            else:
                logger.error("can't assign values, item or section {0} not found in config file!".format(section))


    def configspecParser(self, configspec, comments):
        """
        This is a really trivial parser for ConfigObj spec file. This parser aims to exctract the limits and the available options for the entries in the config file. For example:
        if a config entry is defined as "option('mm', 'in', default = 'mm')", then the parser will create a list with ['mm', 'in]
        similarly, if an entry defined as "integer(max=9)", the max value will be exctracted
        @param configspec: specifications of the configfile. This variable is created by ConfigObj module.
        @param comments: string list containing the comments for a given item
        @return The function returns a dictionary with the following fields
        - minimum : contains the minimum value or length for an entry (possibly 'None' if nothing found)
        - maximum : contains the maximum value or length for an entry (possibly 'None' if nothing found)
        - string_list : contains the list of options for an "option" field, or the column titles for a table
        - comment : a text with the comment that belongs to the parameter (possibly an empty string if nothing found)
        """
        #logger.debug('configspecParser({0}, {1})'.format(configspec, comments))
        minimum = None
        maximum = None
        string_list = []

        if isinstance(configspec, dict):
            #If the received configspec is a dictionary, we most likely have a table, so we are going to exctract sections names of this table

                #When tables are used, the "__many__" config entry is used for the definition of the configspec, so we try to excract the sections names by using this __many__ special keyword.
                #Example: 'Tool_Parameters': {[...], '__many__': {'diameter': 'float(default = 3.0)', 'speed': 'float(default = 6000)', 'start_radius': 'float(default = 3.0)'}}
                if '__many__' in configspec and isinstance(configspec['__many__'], dict):
                    string_list = configspec['__many__'].keys()
                    string_list.insert(0, '') #prepend an empty element since the first column of the table is the row name (eg a unique tool number)

        else:
            #configspec is normaly a string from which we can exctrat min / max values and possibly a list of options

            #Handle "option" config entries
            string_list = self.configspecParserExctractSections('option', configspec)
            i = 0
            while i < len(string_list): #DON'T replace this with a "for", it would silently skip some steps because we remove items inside the loop
                #remove unwanted items which are unquoted (like the "default=" parameter) and remove the quotes
                if string_list[i].startswith('"'):
                    string_list[i] = string_list[i].strip('"')
                elif string_list[i].startswith("'"):
                    string_list[i] = string_list[i].strip("'")
                else:
                    #unwanted item, it doesn't contain an element of the option()
                    del string_list[i]
                    continue

                i += 1

            #Handle "integer" and "string" config entries
            if len(string_list) <= 0:
                string_list = self.configspecParserExctractSections('integer', configspec)
                if len(string_list) <= 0:
                    string_list = self.configspecParserExctractSections('string', configspec)

                minimum, maximum = self.handle_type_config_entries(minimum, maximum, string_list, toInt)

            #Handle "float" config entries
            if len(string_list) <= 0:
                string_list = self.configspecParserExctractSections('float', configspec)

                minimum, maximum = self.handle_type_config_entries(minimum, maximum, string_list, toFloat)

        #Handle comments: comments are stored in a list and contains any chars that are in the configfile (including the hash symbol and the spaces)
        comments_string = ''
        if len(comments) > 0:
            for comment in comments:
                comments_string += comment.strip()

            comments_string = comments_string.strip(' #')
            comments_string = comments_string.replace('#', '\n')

        logger.debug('configspecParser(): exctracted option elements = {0}, min = {1}, max = {2}, comment = {3}'.format(string_list, minimum, maximum, comments_string))

        result = {}
        result['minimum'] = minimum
        result['maximum'] = maximum
        result['string_list'] = string_list
        result['comment'] = comments_string
        return result

    def handle_type_config_entries(self, minimum, maximum, string_list, type_converter):
        for element in string_list:
            if minimum is not None and maximum is not None:
                break

            value = type_converter(element)
            if value[1]:
                if minimum is None:
                    minimum = value[0]
                elif maximum is None:
                    maximum = value[0]

            if minimum is None and 'min' in element:
                # string found in a string like "min = -7"
                element = element.replace('min', '')
                element = element.strip(' =')
                value = type_converter(element)
                if value[1]:
                    minimum = value[0]

            if maximum is None and 'max' in element:
                # 'max' string found
                element = element.replace('max', '')
                element = element.strip(' =')
                value = type_converter(element)
                if value[1]:
                    maximum = value[0]

        return minimum, maximum

    def configspecParserExctractSections(self, attribute_name, string):
        """
        returns a list of item from a string. Eg the string "option('mm', 'in', default = 'mm')" will be exploded into the string list ["mm", "in", "default = 'mm'"]
        """
        string_list = []

        pos_init = string.find(attribute_name + '(')
        if pos_init >= 0:
            pos_init += len(attribute_name + '(') #skip the "option("

            pos_end = string.find(')', pos_init)
            if pos_end > pos_init:
                #print("substring found = {0}".format(string[pos_init:pos_end]))
                string_list = string[pos_init:pos_end].split(',')

        # remove empty elements and remove leading and trailing spaces
        string_list = [string.strip() for string in string_list if string]
        return string_list


    def changeOccured(self):
        """
        This function (slot) is called whenever a modification occurs in the configuration window.
        It enables "Apply" and "OK" buttons, plus disable the configfile selector.
        """
        #There are some changes, we swith to edit mode
        self.setEditInProgress(True)


    def validateConfiguration(self, window_def, result_string = '', result_bool = True):
        """
        Check the configuration (check the limits, eg min/max values, ...). These limits are set according to the configspec passed to the constructor
        @param window_def: the dict that describes our window
        @param result_string: use only for recursive call
        @param result_bool: use only for recursive call
        @return (result_bool, result_string):
         - result_bool: True if no errors were encountered, False otherwise
         - result_string: a string containing all the errors encountered during the validation
        """
        #Compute all the sections
        for section in window_def:
            #skip the special section __section_title__
            if section == '__section_title__' or isinstance(window_def[section], CfgDummy):
                continue

            if isinstance(window_def[section], dict):
                #Browse sublevels
                (result_bool, result_string) = self.validateConfiguration(window_def[section], result_string, result_bool) #Recursive call, until we find a real item (not a dictionnary with subtree)
            else:
                if isinstance(window_def[section], (QWidget, QLayout)):
                    #check that the value is correct for each widget
                    result = window_def[section].validateValue()
                    if result[0] is False: result_bool = False
                    result_string += result[1]
                else:
                    #Item should be a layout or a widget
                    logger.warning("item {0} is not a widget, can't validate it!".format(window_def[section]))

        return (result_bool, result_string)


    def updateConfiguration(self, window_def, config):
        """
        Update the application configuration (ConfigObj) according to the changes made into the ConfigWindow.
        The self.var_dict variable is updated
        @param window_def: the dict that describes our window
        @param config: data readed from the configfile. This dict is created by ConfigObj module and will be updated here.
        """
        #Compute all the sections
        for section in window_def:
            #skip the special section __section_title__
            if section == '__section_title__' or isinstance(window_def[section], CfgDummy):
                continue

            if config is not None and section in config:
                if isinstance(window_def[section], dict):
                    #Browse sublevels
                    self.updateConfiguration(window_def[section], config[section]) #Recursive call, until we find a real item (not a dictionnary with subtree)
                else:
                    if isinstance(window_def[section], (QWidget, QLayout)):
                        #assign the value that was readed from the configfile
                        config[section] = window_def[section].getValue()
                    else:
                        #Item should be a layout or a widget
                        logger.warning("item {0} is not a widget, can't update it!".format(window_def[section]))
            else:
                logger.error("can't update configuration, item or section {0} not found in config file!".format(section))



############################################################################################################################
# The classes below are all based on QWidgets and allow to create various predefined elements for the configuration window #
############################################################################################################################

class CfgBase(QWidget):
    """
    Base class used only for setting the Layout. Want a consistent look.
    """
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

    def setLayout(self, layout=None, indent=True):
        layout.setSpacing(2)  # Don't use too much space, it makes the option window too big otherwise
        layout.setContentsMargins(10 if indent else 0, 5, 1, 1)
        QWidget.setLayout(self, layout)


class CfgDummy(CfgBase):
    """
    If a class inherits this dummy class then it should be skipped - it only serves display purposes
    """
    def __init__(self, parent=None):
        CfgBase.__init__(self, parent)


class CfgSubtitle(CfgDummy):
    def __init__(self, text=None, parent=None):
        CfgDummy.__init__(self, parent)

        layout = QHBoxLayout()

        if text is not None:
            layout.addWidget(QLabel(text))

        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setSizePolicy(QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed))
        layout.addWidget(separator)

        self.setLayout(layout, False)


class CfgCheckBox(CfgBase):
    """
    Subclassed QCheckBox to match our needs.
    """

    def __init__(self, text, tristate = False, parent = None):
        """
        Initialization of the CfgCheckBox class.
        @param text: text string associated with the checkbox
        @param tristate: whether the checkbox must have 3 states (tristate) or 2 states
        """
        CfgBase.__init__(self, parent)

        self.checkbox = QCheckBox(text, parent)
        self.checkbox.setTristate(tristate)

        layout = QHBoxLayout(parent)
        layout.addWidget(self.checkbox)
        self.setLayout(layout)

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        #Nothing is configurable in the configspec for this item
        if spec['comment']:
            self.setWhatsThis(spec['comment'])

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.checkbox.stateChanged.connect(changeNotifyer)

    def validateValue(self):
        """
        This item can't be wrong, so we always return true and an empty string
        @return (True, ''):
        """
        return (True, '')

    def getValue(self):
        """
        @return 0 when the checkbox is unchecked, 1 if it is checked and 2 if it is partly checked (tristate must be set to true for tristate mode)
        """
        check_state = self.checkbox.checkState()
        if check_state == QtCore.Qt.Unchecked:
            check_state = 0
        elif check_state == QtCore.Qt.Checked:
            check_state = 1
        elif check_state == QtCore.Qt.PartiallyChecked:
            check_state = 2

        return check_state if self.checkbox.isTristate() else check_state == 1

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: 0 when the checkbox is unchecked, 1 if it is checked and 2 if it is partly checked (tristate must be set to true for tristate mode)
        """
        self.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        if value == 0:
            self.checkbox.setCheckState(QtCore.Qt.Unchecked)
        elif value == 2:
            self.checkbox.setCheckState(QtCore.Qt.PartiallyChecked)
        else:
            self.checkbox.setCheckState(QtCore.Qt.Checked)
        self.blockSignals(False)


class CfgSpinBox(CfgBase):
    """
    Subclassed QSpinBox to match our needs.
    """

    def __init__(self, text, unit = None, minimum = None, maximum = None, parent = None):
        """
        Initialization of the CfgSpinBox class (used for int values).
        @param text: text string associated with the SpinBox
        @param minimum: min value (int)
        @param minimum: max value (int)
        """
        CfgBase.__init__(self, parent)

        self.label = QLabel(text, parent)

        self.spinbox = QSpinBox(parent)
        self.spinbox.setMinimumWidth(200)  # Provide better alignment with other items

        if unit is not None:
            self.setUnit(unit)

        layout = QHBoxLayout(parent)
        layout.addWidget(self.label)
        layout.addStretch()
        layout.addWidget(self.spinbox)
        self.setLayout(layout)

        self.setSpec({'minimum': minimum, 'maximum': maximum, 'comment': ''})

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        if spec['minimum'] is not None:
            self.spinbox.setMinimum(spec['minimum'])
        else:
            self.spinbox.setMinimum(-1000000000) #if no value is defined for the minimum, use a reasonable value

        if spec['maximum'] is not None:
            self.spinbox.setMaximum(spec['maximum'])
        else:
            self.spinbox.setMaximum(1000000000) #if no value is defined for the maximum, use a more reasonable value than 99 (default value in QT) ...

        if spec['comment']:
            self.setWhatsThis(spec['comment'])

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.spinbox.valueChanged.connect(changeNotifyer)

    def setUnit(self, unit):
        """
        Set the unit of the SpinBox (unit is displayed just after the value)
        @param unit: string with the unit used for the spinbox
        """
        self.spinbox.setSuffix(unit)

    def validateValue(self):
        """
        This item can't be wrong, so we always return true and an empty string
        @return (True, ''):
        """
        return (True, '')

    def getValue(self):
        """
        @return: the current value of the QSpinBox
        """
        return self.spinbox.value()

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: int value
        """
        self.spinbox.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.spinbox.setValue(value)
        self.spinbox.blockSignals(False)


class CorrectedDoubleSpinBox(QDoubleSpinBox):
    """
    Subclassed QDoubleSpinBox to get a version that works for everyone ...
    DON'T remove this class, it is a correction for the guys who decided to use comma (',') as a decimal separator (France, Italy, ...) but failed to use comma on the keypad (keypad use dot, not comma)!
    This subclassed QDoubleSpinBox allow to enter decimal values like 3.5 _and_ 3,5
    (the default QDoubleSpinBox implementation only allow the locale as a decimal separator, so for eg, France, you just can't enter decimal values using the keypad!)
    See here for more details: http://www.qtcentre.org/threads/12483-Validator-for-QDoubleSpinBox and http://www.qtcentre.org/threads/13711-QDoubleSpinBox-dot-as-comma
    """

    def __init__(self, parent = None):
        QDoubleSpinBox.__init__(self, parent)
        self.saved_suffix = ''
        #Let's use the locale decimal separator if it is different from the dot ('.')
        local_decimal_separator = QLocale().decimalPoint()
        if local_decimal_separator == '.':
            local_decimal_separator = ''
        self.lineEdit().setValidator(QRegExpValidator(QRegExp("-?[0-9]*[.{0}]?[0-9]*.*".format(local_decimal_separator)), self))

    def setSuffix(self, suffix):
        self.saved_suffix = suffix
        QDoubleSpinBox.setSuffix(self, suffix)

    def valueFromText(self, text):
        # result = float(text.replace('.', QLocale().decimalPoint()))
        # python expect a dot ('.') as decimal separator
        text = qstr_encode(text).replace(str_encode(self.saved_suffix), '').replace(str(QLocale().decimalPoint()), '.')
        return toFloat(text)[0]

    def validate(self, entry, pos):
        #let's *really* trust the validator
        #http://python.6.x6.nabble.com/QValidator-raises-TypeError-td1923683.html
        #print("validate({}, {})".format(entry, pos))
        return (QValidator.Acceptable, entry, pos)


class CfgDoubleSpinBox(CfgSpinBox):
    """
    Subclassed QDoubleSpinBox to match our needs.
    """

    def __init__(self, text, unit = None, minimum = None, maximum = None, precision = None, parent = None):
        """
        Initialization of the CfgDoubleSpinBox class (used for float values).
        @param text: text string associated with the SpinBox
        @param minimum: min value (float)
        @param minimum: max value (float)
        """
        CfgBase.__init__(self, parent)  # skip the init of CfgSpinBox - we want a "corrected" spinbox

        self.label = QLabel(text, parent)

        self.spinbox = QSpinBox(parent)
        self.spinbox = CorrectedDoubleSpinBox(parent)
        self.spinbox.setMinimumWidth(200)  # Provide better alignment with other items
        if precision is not None:
            self.spinbox.setDecimals(precision)

        if unit is not None:
            self.setUnit(unit)

        layout = QHBoxLayout(parent)
        layout.addWidget(self.label)
        layout.addStretch()
        layout.addWidget(self.spinbox)
        self.setLayout(layout)

        self.setSpec({'minimum': minimum, 'maximum': maximum, 'comment': ''})


class CfgLineEdit(CfgBase):
    """
    Subclassed QLineEdit to match our needs.
    """

    def __init__(self, text, size_min = None, size_max = None, parent = None):
        """
        Initialization of the CfgLineEdit class (text edit, one line).
        @param text: text string associated with the line edit
        @param size_min: min length (int)
        @param size_max: max length (int)
        """
        CfgBase.__init__(self, parent)

        self.label = QLabel(text, parent)

        self.lineedit = QLineEdit(parent)

        layout = QVBoxLayout(parent)
        layout.addWidget(self.label)
        layout.addWidget(self.lineedit)
        self.setLayout(layout)

        self.size_min = 0
        self.setSpec({'minimum': size_min, 'maximum': size_max, 'comment': ''})

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        if spec['minimum'] is not None:
            self.size_min = spec['minimum']

        if spec['maximum'] is not None:
            self.lineedit.setMaxLength(spec['maximum'])

        if spec['comment']:
            self.setWhatsThis(spec['comment'])

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.lineedit.textChanged.connect(changeNotifyer)

    def validateValue(self):
        """
        Check the minimum length value
        @return (result_bool, result_string):
         - result_bool: True if no errors were encountered, False otherwise
         - result_string: a string containing all the errors encountered during the validation
        """
        field_length = len(str(self.lineedit.text()))
        if field_length < self.size_min:
            result = (False, str(self.tr('\nNot enough chars (expected {0}, found {1}) for the field "{2}"\n')).format(self.size_min, field_length, self.label.text()))
        else:
            #OK
            result = (True, '')
        return result

    def getValue(self):
        """
        @return: the current value of the QSpinBox
        """
        return str(self.lineedit.text())

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: text string
        """
        self.lineedit.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.lineedit.setText(value)
        self.lineedit.blockSignals(False)


class CfgListEdit(CfgLineEdit):
    """
    Subclassed QLineEdit to match our needs.
    """

    def __init__(self, text, separator, size_max = None, parent = None):
        """
        Initialization of the CfgListEdit class (text edit, one line, strings separated with "separator").
        @param text: text string associated with the line edit
        @param separator: the separator used for the strings (eg: ',')
        @param size_min: min length (int)
        @param size_max: max length (int)
        """
        CfgLineEdit.__init__(self, text + " (use '" + separator + "' as separator)", size_max, parent)
        #Store the separator so that we can return a list of strings instead of a single string
        self.separator = separator

    def getValue(self):
        """
        @return the current value of the QSpinBox (string list)
        """
        item_list = str(self.lineedit.text()).split(self.separator)
        item_list = [item.strip(' ') for item in item_list]  # remove leading and trailing whitespaces
        return item_list

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: text string or list of text strings
        """
        joined_value = value
        if isinstance(value, (list, tuple)):
            joined_value = (self.separator + ' ').join(value) #Join the strings and add a space for more readability (the space will be removed when writting)

        self.lineedit.setText(joined_value)


class CfgTextEdit(CfgBase):
    """
    Subclassed QTextEdit to match our needs.
    """

    def __init__(self, text, size_min = None, size_max = None, parent = None):
        """
        Initialization of the CfgLineEdit class (text edit, one line).
        @param text: text string associated with the line edit
        @param size_min: min length (int)
        @param size_max: max length (int)
        """
        CfgBase.__init__(self, parent)

        self.label = QLabel(text, parent)

        self.textedit = QTextEdit(parent)
        self.textedit.setAcceptRichText(False)
        self.textedit.setAutoFormatting(QTextEdit.AutoNone)

        layout = QVBoxLayout(parent)
        layout.addWidget(self.label)
        layout.addWidget(self.textedit)
        self.setLayout(layout)

        self.size_min = 0
        self.setSpec({'minimum': size_min, 'maximum': size_max, 'comment': ''})

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        if spec['minimum'] is not None:
            self.size_min = spec['minimum']

        if spec['maximum'] is not None:
            self.textedit.setMaxLength(spec['maximum'])

        if spec['comment']:
            self.setWhatsThis(spec['comment'])

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.textedit.textChanged.connect(changeNotifyer)

    def validateValue(self):
        """
        Check the minimum length value
        @return (result_bool, result_string):
         - result_bool: True if no errors were encountered, False otherwise
         - result_string: a string containing all the errors encountered during the validation
        """
        field_length = len(str(self.textedit.toPlainText()))
        if field_length < self.size_min:
            result = (False, str(self.tr('\nNot enough chars (expected {0}, found {1}) for the field "{2}"\n')).format(self.size_min, field_length, self.label.text()))
        else:
            #OK
            result = (True, '')
        return result

    def getValue(self):
        """
        @return: the current value of the QSpinBox
        """
        return str(self.textedit.toPlainText())

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: text string
        """
        self.textedit.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.textedit.setPlainText(value)
        self.textedit.blockSignals(False)


class CfgComboBox(CfgBase):
    """
    Subclassed QComboBox to match our needs.
    """

    def __init__(self, text, items_list = None, default_item = None, parent = None):
        """
        Initialization of the CfgComboBox class (drop-down menu).
        @param text: text string associated with the combobox
        @param items_list: string list containing all the available options
        @param default_item: string containing the default selected item
        """
        CfgBase.__init__(self, parent)

        if isinstance(items_list, (list, tuple)):
            self.setSpec({'string_list': items_list, 'comment': ''})
        if default_item is not None:
            self.setValue(default_item)

        self.label = QLabel(text, parent)

        self.combobox = QComboBox(parent)
        self.combobox.setMinimumWidth(200)  # Provide better alignment with other items

        layout = QHBoxLayout(parent)
        layout.addWidget(self.label)
        layout.addStretch()
        layout.addWidget(self.combobox)
        self.setLayout(layout)

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        self.combobox.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.combobox.clear()
        self.combobox.addItems(spec['string_list'])

        if spec['comment']:
            self.setWhatsThis(spec['comment'])
        self.combobox.blockSignals(False)

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.combobox.currentIndexChanged.connect(changeNotifyer)

    def validateValue(self):
        """
        This item can't be wrong, so we always return true and an empty string
        @return (True, ''):
        """
        return (True, '')

    def getValue(self):
        """
        @return: the string of the currently selected entry
        """
        return self.combobox.currentText()

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: the text of the entry to select in the combobox
        """
        self.combobox.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.combobox.setCurrentIndex(self.combobox.findText(value)) #Compatible with both PyQt4 and PyQt5
        self.combobox.blockSignals(False)


class CfgTable(QWidget):
    """
    Subclassed QTableWidget to match our needs.
    """
    #Define a QT signal that is emitted when the table is modified.
    #Note: this signal is not emitted in this class ; it is up to the subclasses to emit it
    tableChanged = QtCore.pyqtSignal()

    def __init__(self, text, columns = None, parent = None):
        """
        Initialization of the CfgTable class (editable 2D table).
        @param text: text string associated with the table
        @param columns: string list containing all the columns names
        """
        QWidget.__init__(self, parent)

        self.tablewidget = QTableWidget(parent)
        self.tablewidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        if isinstance(columns, (list, tuple)):
            self.setSpec({'string_list': columns, 'comment': ''})
        else:
            self.keys = [] #No columns yet
        self.tablewidget.horizontalHeader().setStretchLastSection(True)
        self.tablewidget.horizontalHeader().sectionClicked.connect(self.tablewidget.clearSelection) #Allow to unselect the lines by clicking on the column name (useful to add a line at the end)

        self.label = QLabel(text, parent)
        self.button_add = QPushButton(QIcon(QPixmap(":/images/list-add.png")), "")
        self.button_remove = QPushButton(QIcon(QPixmap(":/images/list-remove.png")), "")
        self.button_add.clicked.connect(self.appendLine)
        self.button_remove.clicked.connect(self.removeLine)
        self.layout_button = QVBoxLayout()
        self.layout_button.addWidget(self.button_add)
        self.layout_button.addWidget(self.button_remove)

        self.layout_table = QHBoxLayout()
        #self.tablewidget.setSizePolicy(size_policy)
        self.layout_table.addWidget(self.tablewidget)
        self.layout_table.addLayout(self.layout_button)

        self.layout = QVBoxLayout(parent)

        self.layout.addWidget(self.label)
        self.layout.addLayout(self.layout_table)
        self.setLayout(self.layout)

        #Ensure that the table always expand to the maximum available space
        size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        size_policy.setVerticalStretch(10)
        size_policy.setHorizontalStretch(10)
        self.setSizePolicy(size_policy)

    def setSpec(self, spec):
        """
        Set the specifications for the item (min/max values, ...)
        @param spec: the specifications dict (can contain the following keys: minimum, maximum, comment, string_list)
        """
        self.keys = spec['string_list']
        if len(self.keys) > 0 and not self.keys[0]:
            self.keys[0] = 'name' #name of first column is normaly undefined in configspec, so we use a generic name, just to display something in the header of the QTable
        self.tablewidget.setColumnCount(len(self.keys))
        self.tablewidget.setHorizontalHeaderLabels(self.keys)

        if spec['comment']:
            self.setWhatsThis(spec['comment'])

    def setChangeSlot(self, changeNotifyer):
        """
        Assign a notifyier slot (this slot is called whenever the state of the widget changes)
        @param changeNotifyer: the function (slot) that is called in case of change
        """
        self.tableChanged.connect(changeNotifyer)
        self.tablewidget.itemChanged.connect(changeNotifyer)
        self.tablewidget.cellChanged.connect(changeNotifyer)
        self.button_add.clicked.connect(changeNotifyer)
        self.button_remove.clicked.connect(changeNotifyer)

    def appendLine(self, line = None):
        """
        Add a line to the table. The new line is inserted before the selected line, or at the end of the table is no line is selected
        @param line: a string list containing all the values for this lines. If line is None, an empty line is inserted
        """
        selected_row = self.tablewidget.currentRow()
        if selected_row < 0 or len(self.tablewidget.selectedIndexes()) <= 0: #Trick to be able to insert lines before the first and after the last line (click on column name to unselect the lines)
            selected_row = self.tablewidget.rowCount()

        self.tablewidget.insertRow(selected_row)

        #If provided, fill the table with the content of the line list
        if line is not None and isinstance(line, (list, tuple)) and len(line) >= self.tablewidget.columnCount():
            for i in range(self.tablewidget.columnCount()):
                #self.tablewidget.setItem(selected_row, i, QTableWidgetItem(line[i]))
                self.setCellValue(selected_row, i, line[i])
        else:
            for i in range(self.tablewidget.columnCount()):
                self.setCellValue(selected_row, i, "") #Don't remove this line, otherwise the subclasses won't be able to set custom widget into the table.

        #Resize the columns to the content, except for the last one
        for i in range(self.tablewidget.columnCount() - 1):
            self.tablewidget.resizeColumnToContents(i)

        #Resize the rows to the content
        for i in range(self.tablewidget.rowCount()):
            self.tablewidget.resizeRowToContents(i)

    def removeLine(self):
        """
        Remove a line from the table. The selected line is suppressed, or the last line if no line is selected
        """
        selected_row = self.tablewidget.currentRow()
        if selected_row < 0 and self.tablewidget.rowCount() > 0:
            selected_row = self.tablewidget.rowCount() - 1

        if selected_row >= 0:
            self.tablewidget.removeRow(selected_row)

    def setCellValue(self, line, column, value):
        """
        Default implementation for filling cells use Qt default QTableWidgetItem. One can subclass to provide another implementation, like inserting various Widget into the table.
        @param line: line number (int)
        @param column: column number (int)
        @param value: cell content (string)
        """
        self.tablewidget.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        self.tablewidget.setItem(line, column, QTableWidgetItem(value))
        self.tablewidget.blockSignals(False)

    def validateValue(self):
        """
        Default implementation always return true (OK) and an empty string
        @return (True, ''):
        """
        return (True, '')

    def getValue(self):
        """
        @return: a nested dictionnary that can be directly used to update the ConfigObj (class that handles configuration file)
        Example of returned value:
        {'pause': {'gcode': 'M5 (Spindle off)\nM9 (Coolant off)\nM0\n\nM8\nS5000M03 (Spindle 5000rpm cw)\n'}, 'probe_tool': {'gcode': '\nO<Probe_Tool> CALL\nS6000 M3 M8\n'}}
        """
        result_dict = {}
        for i in range(self.tablewidget.rowCount()):
            key = self.tablewidget.item(i, 0)
            if not key:
                continue
            key = str(key.text())
            if not key:
                continue
            result_dict[key] = {}

            j = 1
            while j < self.tablewidget.columnCount():
                sub_key = self.keys[j] #Column name
                value = self.tablewidget.item(i, j)
                if not value:
                    j += 1
                    continue
                value = str(value.text())
                if sub_key is not None:
                    result_dict[key][sub_key] = value
                j += 1

        return result_dict

    def setValue(self, value):
        """
        Assign the value for our object
        @param value: this is a nested dict, with keys going to the first column of our table and values going to the other columns. Example of received value:
        {'15': {'diameter': 1.5, 'speed': 6000.0, 'start_radius': 1.5}, '20': {'diameter': 2.0, 'speed': 6000.0, 'start_radius': 2.0}, '30': {'diameter': 3.0, 'speed': 6000.0, 'start_radius': 3.0}}
        """
        self.tablewidget.blockSignals(True) # Avoid unnecessary signal (we don't want the config window to emit any signal when filling the fields programatically)
        result = True
        if isinstance(value, dict) and len(self.keys) > 0:
            self.tablewidget.setRowCount(0)
            line = [None] * len(self.keys)

            #sort according to the key
            item_list=[]
            try:
                #try numeric sort
                item_list = sorted(value.keys(), key=float)
            except ValueError:
                #fallback to standard sort
                item_list = sorted(value.keys())

            for item in item_list:
                line[0] = item #First column is alway the key of the dict (eg it can be the tool number)

                #Compute the other columns (the received value must contain a dict entry for each column)
                i = 1
                while i < len(self.keys):
                    if self.keys[i] in value[item]:
                        line[i] = value[item][self.keys[i]] #Get the value for a given column
                    else:
                        result = False
                        break
                    i += 1

                if result is True:
                    self.appendLine(line)
        else:
            result = False

        self.tablewidget.blockSignals(False)

        return result


class CfgTableCustomActions(CfgTable):
    """
    Subclassed CfgTableWidget to use muli-line edits for storing the custom GCODE.
    """

    def __init__(self, text, columns = None, parent = None):
        """
        Initialization of the CfgTableCustomActions class (editable 2D table for storing custom GCODE).
        @param text: text string associated with the table
        @param columns: string list containing all the columns names
        """
        CfgTable.__init__(self, text, columns, parent)

    def setCellValue(self, line, column, value):
        """
        This function is reimplemented to use QTextEdit into the table, thus allowing multi-lines custom GCODE to be stored
        @param line: line number (int)
        @param column: column number (int)
        @param value: cell content (string)
        """
        if column > 0:
            #Special case for column 1 : we use QTextEdit for storing the GCODE
            text_edit = QTextEdit()
            text_edit.setAcceptRichText(False)
            text_edit.setAutoFormatting(QTextEdit.AutoNone)
            text_edit.setPlainText(value)
            self.tablewidget.setCellWidget(line, column, text_edit)
            text_edit.textChanged.connect(self.valueChangedSlot) #We need to detect any change in the table's widgets
        else:
            #Normal case: use standard QT functions (and we dont need to generate any signal here, it is already handled by the parent class)
            CfgTable.setCellValue(self, line, column, value)

    def valueChangedSlot(self):
        """
        Slot called when something changes in the table. We emit a signal, so that the config window can detect this change
        """
        self.tableChanged.emit()

    def validateValue(self):
        """
        Check that the keys are unique and not empty
        @return (result_bool, result_string):
         - result_bool: True if no errors were encountered, False otherwise
         - result_string: a string containing all the errors encountered during the validation
        """
        #For now everything is OK
        result_string = ''
        result_bool = True
        keys_list = []

        if self.tablewidget.rowCount() > 0 and self.tablewidget.columnCount() > 0:
            for i in range(self.tablewidget.rowCount()):
                if not self.tablewidget.item(i, 0) or not self.tablewidget.item(i, 0).text():
                    result_bool = False
                    result_string += str(self.tr('\nThe cell at line {0}, column 0 must not be empty for the table "{1}"\n')).format(i, self.label.text())
                else:
                    #Create a list with all the "keys" from the first column (here a key is the custom action name)
                    keys_list.append(self.tablewidget.item(i, 0).text())

            nb_duplicate_elements = len(keys_list) - len(set(keys_list))
            if nb_duplicate_elements != 0:
                #There are duplicate entries, that's wrong because the key must be unique
                result_bool = False
                result_string += str(self.tr('\nFound {0} duplicate elements for the table "{1}"\n')).format(nb_duplicate_elements, self.label.text())

        return (result_bool, result_string)

    def getValue(self):
        """
        @return: a nested dictionnary that can be directly used to update the ConfigObj (class that handles configuration file)
        Example of returned value:
        {'pause': {'gcode': 'M5 (Spindle off)\nM9 (Coolant off)\nM0\n\nM8\nS5000M03 (Spindle 5000rpm cw)\n'}, 'probe_tool': {'gcode': '\nO<Probe_Tool> CALL\nS6000 M3 M8\n'}}
        """
        result_dict = {}
        #Get the keys (first column)
        for i in range(self.tablewidget.rowCount()):
            key = self.tablewidget.item(i, 0)
            if not key:
                continue
            key = str(key.text())
            if not key:
                continue
            result_dict[key] = {}

            #Get the values (other columns)
            for j in range(1, self.tablewidget.columnCount()):
                sub_key = self.keys[j] #Column name
                value = self.tablewidget.cellWidget(i, j)
                if not value:
                    continue
                value = str(value.toPlainText())
                if sub_key is not None:
                    result_dict[key][sub_key] = value

        return result_dict


class CfgTableToolParameters(CfgTable):
    """
    Subclassed CfgTableWidget to use muli-line edits for storing the custom GCODE.
    """

    def __init__(self, text, columns = None, parent = None):
        """
        Initialization of the CfgTableWidget class (editable 2D table for storing the tools table).
        @param text: text string associated with the table
        @param columns: string list containing all the columns names
        """
        self.max_tool_number = 0
        CfgTable.__init__(self, text, columns, parent)

    def setCellValue(self, line, column, value):
        """
        This function is reimplemented to use QTextEdit into the table, thus allowing multi-lines custom GCODE to be stored
        @param line: line number (int)
        @param column: column number (int)
        @param value: cell content (string or int or float)
        """
        if column > 0:
            #we use QDoubleSpinBox for storing the values
            spinbox = CorrectedDoubleSpinBox()
            spinbox.setMinimum(0)
            spinbox.setMaximum(1000000000) #Default value is 99
            computed_value = toFloat(value)[0]  # Convert the value to float
            spinbox.setValue(computed_value)
        else:
            #tool number is an integer
            spinbox = QSpinBox()
            spinbox.setMinimum(0)
            spinbox.setMaximum(1000000000) #Default value is 99

            converted_value = toInt(value)  # Convert the value to int (we may receive a string for example)
            computed_value = converted_value[0] if converted_value[1] else self.max_tool_number + 1
            self.max_tool_number = max(self.max_tool_number, computed_value) #Store the max value for the tool number, so that we can automatically increment this value for new tools
            spinbox.setValue(computed_value) #first column is the key, it must be an int

        spinbox.valueChanged.connect(self.valueChangedSlot) #We need to detect any change in the table's widgets

        self.tablewidget.setCellWidget(line, column, spinbox)

    def valueChangedSlot(self):
        """
        Slot called when something changes in the table. We emit a signal, so that the config window can detect this change
        """
        self.tableChanged.emit()

    def validateValue(self):
        """
        Check that the keys are unique and not empty
        @return (result_bool, result_string):
         - result_bool: True if no errors were encountered, False otherwise
         - result_string: a string containing all the errors encountered during the validation
        """
        #For now everything is OK
        contains_tool_1 = False
        result_string = ''
        result_bool = True
        keys_list = []

        if self.tablewidget.rowCount() > 0 and self.tablewidget.columnCount() > 0:
            for i in range(self.tablewidget.rowCount()):
                if not self.tablewidget.cellWidget(i, 0):
                    result_bool = False
                    result_string += str(self.tr('\nThe cell at line {0}, column 0 must not be empty for the table "{1}"\n')).format(i, self.label.text())
                else:
                    #Create a list with all the "keys" from the first column (here a key is the custom action name)
                    keys_list.append(str(self.tablewidget.cellWidget(i, 0).value()))
                    if self.tablewidget.cellWidget(i, 0).value() == 1:
                        contains_tool_1 = True

            nb_duplicate_elements = len(keys_list) - len(set(keys_list))
            if nb_duplicate_elements != 0:
                #There are duplicate entries, that's wrong because the key must be unique
                result_bool = False
                result_string += str(self.tr('\nFound {0} duplicate elements for the table "{1}"\n')).format(nb_duplicate_elements, self.label.text())

        if not contains_tool_1:
            result_bool = False
            result_string += str(self.tr('\nThe table "{0}" must always contains tool number \'1\'\n')).format(self.label.text()) #Note: str() is needed for PyQt4

        return (result_bool, result_string)

    def getValue(self):
        """
        @return: a nested dictionnary that can be directly used to update the ConfigObj (class that handles configuration file)
        Example of returned value:
        {'pause': {'gcode': 'M5 (Spindle off)\nM9 (Coolant off)\nM0\n\nM8\nS5000M03 (Spindle 5000rpm cw)\n'}, 'probe_tool': {'gcode': '\nO<Probe_Tool> CALL\nS6000 M3 M8\n'}}
        """
        result_dict = {}
        #Get the keys (first column)
        for i in range(self.tablewidget.rowCount()):
            key = self.tablewidget.cellWidget(i, 0)
            if not key:
                continue
            key = str(key.value())
            if not key:
                continue
            result_dict[key] = {}

            #Get the values (other columns)
            for j in range(1,self.tablewidget.columnCount()):
                sub_key = self.keys[j] #Column name
                value = self.tablewidget.cellWidget(i, j)
                if not value:
                    continue
                value = value.value()
                if sub_key is not None:
                    result_dict[key][sub_key] = value

        return result_dict



# -*- coding: utf-8 -*-

############################################################################
#
#   Copyright (C) 2008-2015
#    Christian Kohlöffel
#    Vinzenz Schulz
#    Jean-Paul Schouwstra
#
#   This file is part of DXF2GCODE.
#
#   DXF2GCODE is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   DXF2GCODE is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with DXF2GCODE.  If not, see <http://www.gnu.org/licenses/>.
#
############################################################################

from __future__ import division

from math import sqrt, pi
from copy import deepcopy

from dxf2gcode.core.point import Point
from dxf2gcode.core.boundingbox import BoundingBox
#from dxf2gcode.core.arcgeo import ArcGeo

import logging
logger = logging.getLogger("core.linegeo")

eps = 1e-12


class LineGeo(object):

    """
    Standard Geometry Item used for DXF Import of all geometries, plotting and
    G-Code export.
    """

    def __init__(self, Ps, Pe):
        """
        Standard Method to initialize the LineGeo.
        @param Ps: The Start Point of the line
        @param Pe: the End Point of the line
        """
        self.Ps = Ps
        self.Pe = Pe
        self.length = self.Ps.distance(self.Pe)

        self.calc_bounding_box()

        self.abs_geo = None

    def __deepcopy__(self, memo):
        return LineGeo(deepcopy(self.Ps, memo),
                       deepcopy(self.Pe, memo))

    def __str__(self):
        """
        Standard method to print the object
        @return: A string
        """
        return ("\nLineGeo(Ps=Point(x=%s ,y=%s),\n" % (self.Ps.x, self.Ps.y)) + \
               ("Pe=Point(x=%s, y=%s))" % (self.Pe.x, self.Pe.y))

    def save_v1(self):
        return "\nLineGeo" +\
               "\nPs:     %s" % self.Ps.save_v1() +\
               "\nPe:     %s" % self.Pe.save_v1() +\
               "\nlength: %0.5f" % self.length

    def calc_bounding_box(self):
        """
        Calculated the BoundingBox of the geometry and saves it into self.BB
        """
        Ps = Point(x=min(self.Ps.x, self.Pe.x), y=min(self.Ps.y, self.Pe.y))
        Pe = Point(x=max(self.Ps.x, self.Pe.x), y=max(self.Ps.y, self.Pe.y))

        self.BB = BoundingBox(Ps=Ps, Pe=Pe)

    def get_start_end_points(self, start_point, angles=None):
        if start_point:
            if angles is None:
                return self.Ps
            elif angles:
                return self.Ps, self.Ps.norm_angle(self.Pe)
            else:
                return self.Ps, (self.Pe - self.Ps).unit_vector()
        else:
            if angles is None:
                return self.Pe
            elif angles:
                return self.Pe, self.Pe.norm_angle(self.Ps)
            else:
                return self.Pe, (self.Pe - self.Ps).unit_vector()

    def distance_l_p(self, Point):
        """
        Find the shortest distance between CCLineGeo and Point elements.
        Algorithm acc. to
        http://notejot.com/2008/09/distance-from-Point-to-line-segment-in-2d/
        http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm
        @param Point: the Point
        @return: The shortest distance between the Point and Line
        """
        d = self.Pe - self.Ps
        v = Point - self.Ps

        t = d.dotProd(v)

        if t <= 0:
            # our Point is lying "behind" the segment
            # so end Point 1 is closest to Point and distance is length of
            # vector from end Point 1 to Point.
            return self.Ps.distance(Point)
        elif t >= d.dotProd(d):
            # our Point is lying "ahead" of the segment
            # so end Point 2 is closest to Point and distance is length of
            # vector from end Point 2 to Point.
            return self.Pe.distance(Point)
        else:
            # our Point is lying "inside" the segment
            # i.e.:a perpendicular from it to the line that contains the line
            if (v.dotProd(v) - (t * t) / d.dotProd(d)) < eps:
                return 0.0
            else:
                return sqrt(v.dotProd(v) - (t * t) / d.dotProd(d))

    def isHit(self, caller, xy, tol):
        """
        This function returns true if the nearest point between the two geometries is within the square of the 
        given tolerance
        @param caller: This is the calling entities (only used in holegeo)
        @param xy: The point which shall be used to determine the distance
        @tol: The tolerance which is used for Hit testing.
        """
        return self.distance_l_p(xy) <= tol

    def make_abs_geo(self, parent=None):
        """
        Generates the absolute geometry based on itself and the parent. This
        is done for rotating and scaling purposes
        """
        Ps = self.Ps.rot_sca_abs(parent=parent)
        Pe = self.Pe.rot_sca_abs(parent=parent)

        self.abs_geo = LineGeo(Ps=Ps, Pe=Pe)

    def make_path(self, caller, drawHorLine):
        drawHorLine(caller, self.Ps, self.Pe)

    def reverse(self):
        """
        Reverses the direction of the arc (switch direction).
        """
        self.Ps, self.Pe = self.Pe, self.Ps
        if self.abs_geo:
            self.abs_geo.reverse()

    def to_short_string(self):
        return "(%f, %f) -> (%f, %f)" % (self.Ps.x, self.Ps.y, self.Pe.x, self.Pe.y)

    def update_start_end_points(self, start_point, value):
        prv_ang = self.Ps.norm_angle(self.Pe)
        if start_point:
            self.Ps = value
        else:
            self.Pe = value
        new_ang = self.Ps.norm_angle(self.Pe)

        if 2 * abs(((prv_ang - new_ang) + pi) % (2 * pi) - pi) >= pi:
            # seems very unlikely that this is what you want - the direction
            # changed (too drastically)
            self.Ps, self.Pe = self.Pe, self.Ps

        self.length = self.Ps.distance(self.Pe)

    def Write_GCode(self, PostPro):
        """
        Writes the GCODE for a Line.
        @param PostPro: The PostProcessor instance to be used
        @return: Returns the string to be written to a file.
        """
        Ps = self.get_start_end_points(True)
        Pe = self.get_start_end_points(False)
        return PostPro.lin_pol_xy(Ps, Pe)
